@@ -14,6 +14,7 @@
 #include <assert.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <inttypes.h>
 #include <netinet/in.h>
 #include <poll.h>
 #include <stdlib.h>
@@ -23,20 +24,18 @@
 #include <unistd.h>
 
 #include <sstream>
+#include <vector>
 
-#include "dbg.h"
+#include "log.h"
+#include "session.h"
+#include "types.h"
 
 #define INTERRUPT_CHAR '\x03'
 
 #ifdef DEBUGTAG
-# define unhandled_req(_g, _m, ...)		\
-	fatal(_m, ## __VA_ARGS__)
+# define UNHANDLED_REQ(_g) FATAL()
 #else
-# define unhandled_req(_g, _m, ...)		\
-	do {					\
-		log_info(_m, ##__VA_ARGS__);	\
-		write_packet(_g, "");		\
-	} while (0)
+# define UNHANDLED_REQ(_g) write_packet(_g, ""); LOG(info)
 #endif
 
 using namespace std;
@@ -44,18 +43,8 @@ using namespace std;
 /**
  * This struct wraps up the state of the gdb protocol, so that we can
  * offer a (mostly) stateless interface to clients.
- *
- * dbg_contexts are mapped into their own (unlinked) shmem.  We do
- * this in order to save/restore debugger state across replay
- * restarts, which is pretty tricky do otherwise.  Before replay
- * restart, we save a cookie in the environment that lets us find the
- * fd of our mapping.  Then after exec(), we re-init our state by
- * finding the cookie and remapping the region (then unsetting the
- * cookie, of course).
  */
 struct dbg_context {
-	// fd of dbg_context shmem region.
-	int desc_fd;
 	// Current request to be processed.
 	struct dbg_request req;
 	// Thread to be resumed.
@@ -91,7 +80,6 @@ bool dbg_is_resume_request(const struct dbg_request* req)
 	switch (req->type) {
 	case DREQ_CONTINUE:
 	case DREQ_STEP:
-	case DREQ_DETACH:
 		return true;
 	default:
 		return false;
@@ -114,42 +102,16 @@ static void make_cloexec(int fd)
 {
 	int flags = fcntl(fd, F_GETFD);
 	if (-1 == flags) {
-		fatal("Can't GETFD flags");
+		FATAL() << "Can't GETFD flags";
 	}
 	if (fcntl(fd, F_SETFD, flags | FD_CLOEXEC)) {
-		fatal("Can't make client socket CLOEXEC");
+		FATAL() << "Can't make client socket CLOEXEC";
 	}
 }
 
-static size_t dbg_context_mapped_size()
-{
-	return ceil_page_size(sizeof(struct dbg_context));
-}
-
-static dbg_context* map_dbg_context(int desc_fd)
-{
-	void* m = mmap(nullptr, dbg_context_mapped_size(),
-		       PROT_READ | PROT_WRITE, MAP_SHARED, desc_fd, 0);
-	if ((void*)-1 == m) {
-		fatal("Couldn't mmap dbg_context desc_fd %d", desc_fd);
-	}
-	debug("mmap dbg_context desc_fd %d at %p", desc_fd, m);
-	return reinterpret_cast<struct dbg_context*>(m);
-}
-
 static dbg_context* new_dbg_context()
 {
-	static int counter = 0;
-	stringstream ss;
-	ss << "rr-dbg-ctx-" << getpid() << "-" << ++counter;
-	// NB: we don't set CLOEXEC here because we may need to
-	// preserve the backing segment across exec(), if replay
-	// restarts.
-	int desc_fd = create_shmem_segment(ss.str().c_str(),
-					   dbg_context_mapped_size(),
-					   O_NO_CLOEXEC);
-	struct dbg_context* dbg = map_dbg_context(desc_fd);
-	dbg->desc_fd = desc_fd;
+	struct dbg_context* dbg = new dbg_context();
 	dbg->insize = sizeof(dbg->inbuf);
 	dbg->outsize = sizeof(dbg->outbuf);
 	return dbg;
@@ -162,67 +124,41 @@ static void open_socket(struct dbg_context* dbg,
 	int ret;
 
 	dbg->listen_fd = socket(AF_INET, SOCK_STREAM, 0);
+	if (dbg->listen_fd < 0) {
+		FATAL() << "Couldn't create socket";
+	}
 	make_cloexec(dbg->listen_fd);
 
 	dbg->addr.sin_family = AF_INET;
 	dbg->addr.sin_addr.s_addr = inet_addr(address);
 	reuseaddr = 1;
-	setsockopt(dbg->listen_fd, SOL_SOCKET, SO_REUSEADDR,
-		   &reuseaddr, sizeof(reuseaddr));
+	ret = setsockopt(dbg->listen_fd, SOL_SOCKET, SO_REUSEADDR,
+			 &reuseaddr, sizeof(reuseaddr));
+	if (ret < 0) {
+		FATAL() << "Couldn't set SO_REUSEADDR";
+	}
+
 	do {
 		dbg->addr.sin_port = htons(port);
 		ret = ::bind(dbg->listen_fd,
 			     (struct sockaddr*)&dbg->addr, sizeof(dbg->addr));
-		if (ret && (EADDRINUSE == errno || EACCES == errno)) {
+		if (ret && (EADDRINUSE == errno || EACCES == errno
+			    || EINVAL == errno)) {
 			continue;
 		}
-		if (ret != 0) {
-			break;
+		if (ret) {
+			FATAL() << "Couldn't bind to port " << port;
 		}
 
 		ret = listen(dbg->listen_fd, 1/*backlogged connection*/);
-		if (ret == 0 || EADDRINUSE != errno) {
-			break;
+		if (ret && EADDRINUSE == errno) {
+			continue;
 		}
+		if (ret) {
+			FATAL() << "Couldn't listen on port " << port;
+		}
+		break;
 	} while (++port, probe);
-	if (ret) {
-		fatal("Couldn't bind to port %d", port);
-	}
-}
-
-static string make_exec_cookie_envvar()
-{
-	stringstream ss;
-	ss << "_RR_DBG_CONTEXT_EXEC_COOKIE_" << getpid();
-	return ss.str();
-}
-
-/**
- * If an exec cookie is found, restore and return the context it
- * points at.  Otherwise return nullptr.
- */
-static struct dbg_context* try_load_from_exec_cookie()
-{
-	string envvar = make_exec_cookie_envvar();
-	const char* cookiep = getenv(envvar.c_str());
-	if (!cookiep) {
-		return nullptr;
-	}
-	string cookie = cookiep;
-	unsetenv(envvar.c_str());
-
-	assert('\0' != cookie[0]);
-	int desc_fd = stoi(cookie);
-	debug("found exec cookie %s=%d", envvar.c_str(), desc_fd);
-
-	struct dbg_context* dbg = map_dbg_context(desc_fd);
-	// Sanity-check the restored mapping.
-	assert(dbg->desc_fd == desc_fd);
-	assert(dbg->listen_fd == -1);
-	assert(dbg->sock_fd >= 0);
-	assert(dbg->insize == sizeof(dbg->inbuf));
-	assert(dbg->outsize == sizeof(dbg->outbuf));
-	return dbg;
 }
 
 /**
@@ -239,7 +175,7 @@ static void await_debugger(struct dbg_context* dbg)
 	// We might restart this debugging session, so don't set the
 	// socket fd CLOEXEC.
 	if (fcntl(dbg->sock_fd, F_SETFL, O_NONBLOCK)) {
-		fatal("Can't make client socket NONBLOCK");
+		FATAL() <<"Can't make client socket NONBLOCK";
 	}
 	close(dbg->listen_fd);
 	dbg->listen_fd = -1;
@@ -259,13 +195,6 @@ struct dbg_context* dbg_await_client_connection(const char* addr,
 						pid_t client,
 						int client_params_fd)
 {
-	if (struct dbg_context* dbg = try_load_from_exec_cookie()) {
-		assert(tgid == dbg->tgid);
-		// Replay just restarted.  |dbg| is now restored to
-		// what it was in the prior session.
-		return dbg;
-	}
-
 	struct dbg_context* dbg = new_dbg_context();
 	open_socket(dbg, addr, desired_port, probe);
 	int port = ntohs(dbg->addr.sin_port);
@@ -294,12 +223,31 @@ struct dbg_context* dbg_await_client_connection(const char* addr,
 			!strcmp(addr, "127.0.0.1") ? "" : addr, port);
 	}
 	dbg->tgid = tgid;
-	debug("limiting debugger traffic to tgid %d", tgid);
+	LOG(debug) <<"limiting debugger traffic to tgid "<< tgid;
 	await_debugger(dbg);
 	return dbg;
 }
 
-void dbg_launch_debugger(int params_pipe_fd)
+static string create_gdb_command_file(const char* macros)
+{
+	char tmp[] = "/tmp/rr-gdb-commands-XXXXXX";
+	// This fd is just leaked. That's fine since we only call this once
+	// per rr invocation at the moment.
+	int fd = mkstemp(tmp);
+	unlink(tmp);
+
+	ssize_t len = strlen(macros);
+	int written = write(fd, macros, len);
+	if (written != len) {
+		FATAL() <<"Failed to write gdb command file";
+	}
+
+	stringstream procfile;
+	procfile << "/proc/" << getpid() << "/fd/" << fd;
+	return procfile.str();
+}
+
+void dbg_launch_debugger(int params_pipe_fd, const char* macros)
 {
 	struct debugger_params params;
 	ssize_t nread = read(params_pipe_fd, &params, sizeof(params));
@@ -308,39 +256,47 @@ void dbg_launch_debugger(int params_pipe_fd)
 	stringstream attach_cmd;
 	attach_cmd << "target extended-remote " << params.socket_addr << ":"
 		   << params.port;
-	debug("launching gdb with command '%s'", attach_cmd.str().c_str());
-	execlp("gdb", "gdb",
-	       // The gdb protocol uses the "vRun" packet to reload
-	       // remote targets.  The packet is specified to be like
-	       // "vCont", in which gdb waits infinitely long for a
-	       // stop reply packet.  But in practice, gdb client
-	       // expects the vRun to complete within the remote-reply
-	       // timeout, after which it issues vCont.  The timeout
-	       // causes gdb<-->rr communication to go haywire.
-	       //
-	       // rr can take a very long time indeed to send the
-	       // stop-reply to gdb after restarting replay; the time
-	       // to reach a specified execution target is
-	       // theoretically unbounded.  Timing our on vRun is
-	       // technically a gdb bug, but because the rr replay and
-	       // the gdb reload models don't quite match up, we'll
-	       // work around it on the rr side by disabling the
-	       // remote-reply timeout.
-	       "-l", "-1",
-	       params.exe_image,
-	       "-ex", attach_cmd.str().c_str(), NULL);
-	fatal("Failed to exec gdb.");
-}
+	LOG(debug) <<"launching gdb with command '"<< attach_cmd.str() <<"'";
 
-void dbg_prepare_restore_after_exec_restart(struct dbg_context* dbg)
-{
-	string envvar = make_exec_cookie_envvar();
-	assert(!getenv(envvar.c_str()));
-
-	stringstream ss;
-	ss << dbg->desc_fd;
-	setenv(envvar.c_str(), ss.str().c_str(), 1/*override*/);
-	debug("set exec cookie to %s=%s", envvar.c_str(), ss.str().c_str());
+	vector<const char*> args;
+	args.push_back("gdb");
+	// The gdb protocol uses the "vRun" packet to reload
+	// remote targets.  The packet is specified to be like
+	// "vCont", in which gdb waits infinitely long for a
+	// stop reply packet.  But in practice, gdb client
+	// expects the vRun to complete within the remote-reply
+	// timeout, after which it issues vCont.  The timeout
+	// causes gdb<-->rr communication to go haywire.
+	//
+	// rr can take a very long time indeed to send the
+	// stop-reply to gdb after restarting replay; the time
+	// to reach a specified execution target is
+	// theoretically unbounded.  Timing our on vRun is
+	// technically a gdb bug, but because the rr replay and
+	// the gdb reload models don't quite match up, we'll
+	// work around it on the rr side by disabling the
+	// remote-reply timeout.
+	args.push_back("-l");
+	args.push_back("-1");
+	args.push_back(params.exe_image);
+	const string& gdb_command_file_path =
+		rr_flags()->gdb_command_file_path;
+	if (gdb_command_file_path.length() > 0 ) {
+		args.push_back("-x");
+		args.push_back(gdb_command_file_path.c_str());
+	}
+	args.push_back("-l");
+	args.push_back("-1");
+	if (macros) {
+		string gdb_command_file = create_gdb_command_file(macros);
+		args.push_back("-x");
+		args.push_back(gdb_command_file.c_str());
+	}
+	args.push_back("-ex");
+	args.push_back(attach_cmd.str().c_str());
+	args.push_back(nullptr);
+	execvp("gdb", (char* const*)args.data());
+	FATAL() <<"Failed to exec gdb.";
 }
 
 /**
@@ -360,7 +316,7 @@ static int poll_socket(const struct dbg_context* dbg,
 
 	ret = poll(&pfd, 1, timeoutMs);
 	if (ret < 0) {
-		fatal("Polling gdb socket failed");
+		FATAL() <<"Polling gdb socket failed";
 	}
 	return ret;
 }
@@ -387,11 +343,12 @@ static void read_data_once(struct dbg_context* dbg)
 	nread = read(dbg->sock_fd, dbg->inbuf + dbg->inlen,
 		     dbg->insize - dbg->inlen);
 	if (0 == nread) {
-		log_info("(gdb closed debugging socket, exiting)");
+		LOG(info) <<"(gdb closed debugging socket, exiting)";
+		dbg_destroy_context(&dbg);
 		exit(0);
 	}
 	if (nread <= 0) {
-		fatal("Error reading from gdb");
+		FATAL() <<"Error reading from gdb";
 	}
 	dbg->inlen += nread;
 	assert("Impl dynamic alloc if this fails (or double inbuf size)"
@@ -407,7 +364,7 @@ static void write_flush(struct dbg_context* dbg)
 
 #ifdef DEBUGTAG
 	dbg->outbuf[dbg->outlen] = '\0';
-	debug("write_flush: '%s'", dbg->outbuf);
+	LOG(debug) <<"write_flush: '"<< dbg->outbuf <<"'";
 #endif
 	while (write_index < dbg->outlen) {
 		ssize_t nwritten;
@@ -417,7 +374,7 @@ static void write_flush(struct dbg_context* dbg)
 				 dbg->outbuf + write_index,
 				 dbg->outlen - write_index);
 		if (nwritten < 0) {
-			fatal("Error writing to gdb");
+			FATAL() <<"Error writing to gdb";
 		}
 		write_index += nwritten;
 	}
@@ -491,7 +448,7 @@ static void write_binary_packet(struct dbg_context* dbg, const char* pfx,
 		}
 	}
 
-	debug(" ***** NOTE: writing binary data, upcoming debug output may be truncated");
+	LOG(debug) <<" ***** NOTE: writing binary data, upcoming debug output may be truncated";
 	return write_packet_bytes(dbg, buf, buf_num_bytes);
 }
 
@@ -631,6 +588,18 @@ static void read_packet(struct dbg_context* dbg)
 	}
 }
 
+static void read_binary_data(const byte* payload, ssize_t data_len, byte* data)
+{
+	ssize_t payload_idx = 0;
+	for (ssize_t i = 0; i < data_len; ++i) {
+		byte b = payload[payload_idx++];
+		if ('}' == b) {
+			b = 0x20 ^ payload[payload_idx++];
+		}
+		data[i] = b;
+	}
+}
+
 /**
  * Parse and return a gdb thread-id from |str|.  |endptr| points to
  * the character just after the last character in the thread-id.  It
@@ -662,7 +631,7 @@ static dbg_threadid_t parse_threadid(const char* str, char** endptr)
 
 static int xfer(struct dbg_context* dbg, const char* name, char* args)
 {
-	debug("gdb asks us to transfer %s(%s)", name, args);
+	LOG(debug) <<"gdb asks us to transfer "<< name <<"("<< args <<")";
 
 	if (!strcmp(name, "auxv")) {
 		assert(!strncmp(args, "read::",	sizeof("read::") - 1));
@@ -671,11 +640,85 @@ static int xfer(struct dbg_context* dbg, const char* name, char* args)
 		dbg->req.target = dbg->query_thread;
 		return 1;
 	}
+	if (name == strstr(name, "siginfo")) {
+		if (args == strstr(args, "read")) {
+			dbg->req.type = DREQ_READ_SIGINFO;
+			args += strlen("read");
+			assert(':' == *args++);
+			assert(':' == *args++);
 
-	unhandled_req(dbg, "Unhandled gdb xfer request: %s(%s)", name, args);
+			dbg->req.mem.addr = (void*)strtoul(args, &args, 16);
+			assert(',' == *args++);
+			
+			dbg->req.mem.len = strtoul(args, &args, 16);
+			assert('\0' == *args);
+
+			return 1;
+		}
+		if (args == strstr(args, "write")) {
+			dbg->req.type = DREQ_WRITE_SIGINFO;
+			return 1;
+		}
+		UNHANDLED_REQ(dbg) <<"Unhandled 'siginfo' request: "<< args;
+		return 0;
+	}
+
+	UNHANDLED_REQ(dbg) <<"Unhandled gdb xfer request: "<< name <<"("<< args <<")";
 	return 0;
 }
 
+/**
+ * Format |value| into |buf| in the manner gdb expects.  |buf| must
+ * point at a buffer with at least |1 + 2*DBG_MAX_REG_SIZE| bytes
+ * available.  Fewer bytes than that may be written, but |buf| is
+ * guaranteed to be null-terminated.
+ */
+static size_t print_reg_value(const DbgRegister& reg, char* buf) {
+	assert(reg.size <= DBG_MAX_REG_SIZE);
+	if (reg.defined) {
+		/* gdb wants the register value in native endianness.
+		 * reg.value read in native endianness is exactly that.
+		 */
+		for (size_t i = 0; i < reg.size; ++i) {
+			snprintf(&buf[2 * i], 3, "%02lx", (unsigned long)reg.value[i]);
+		}
+	} else {
+		for (size_t i = 0; i < reg.size; ++i) {
+			strcpy(&buf[2 * i], "xx");
+		}
+	}
+	return reg.size * 2;
+}
+
+/**
+ * Read the encoded register value in |strp| into |reg|.  |strp| may
+ * be mutated.
+ */
+static void read_reg_value(char** strp, DbgRegister* reg)
+{
+	char* str = *strp;
+
+	if ('x' == str[0]) {
+		reg->defined = false;
+		reg->size = 0;
+		return;
+	}
+
+	reg->defined = true;
+	reg->size = strlen(str) / 2;
+	for (size_t i = 0; i < reg->size; ++i) {
+		char tmp = str[2];
+		str[2] = '\0';
+
+		reg->value[i] = strtoul(str, &str, 16);
+		assert('\0' == *str);
+
+		str[0] = tmp;
+	}
+
+	*strp = str;
+}
+
 static int query(struct dbg_context* dbg, char* payload)
 {
 	const char* name;
@@ -688,19 +731,19 @@ static int query(struct dbg_context* dbg, char* payload)
 	name = payload;
 
 	if (!strcmp(name, "C")) {
-		debug("gdb requests current thread ID");
+		LOG(debug) <<"gdb requests current thread ID";
 		dbg->req.type = DREQ_GET_CURRENT_THREAD;
 		return 1;
 	}
 	if (!strcmp(name, "Attached")) {
-		debug("gdb asks if this is a new or existing process");
+		LOG(debug) <<"gdb asks if this is a new or existing process";
 		/* Tell gdb this is an existing process; it might be
 		 * (see emergency_debug()). */
 		write_packet(dbg, "1");
 		return 0;
 	}
 	if (!strcmp(name, "fThreadInfo")) {
-		debug("gdb asks for thread list");
+		LOG(debug) <<"gdb asks for thread list";
 		dbg->req.type = DREQ_GET_THREAD_LIST;
 		return 1;
 	}
@@ -709,13 +752,13 @@ static int query(struct dbg_context* dbg, char* payload)
 		return 0;
 	}
 	if (!strcmp(name, "GetTLSAddr")) {
-		debug("gdb asks for TLS addr");
+		LOG(debug) <<"gdb asks for TLS addr";
 		/* TODO */
 		write_packet(dbg, "");
 		return 0;
 	}
 	if (!strcmp(name, "Offsets")) {
-		debug("gdb asks for section offsets");
+		LOG(debug) <<"gdb asks for section offsets";
 		dbg->req.type = DREQ_GET_OFFSETS;
 		dbg->req.target = dbg->query_thread;
 		return 1;
@@ -729,17 +772,18 @@ static int query(struct dbg_context* dbg, char* payload)
 		char supported[1024];
 
 		/* TODO process these */
-		debug("gdb supports %s", args);
+		LOG(debug) <<"gdb supports "<< args;
 
 		snprintf(supported, sizeof(supported) - 1,
-			 "PacketSize=%x;QStartNoAckMode+;qXfer:auxv:read+"
+			 "PacketSize=%zd;QStartNoAckMode+;qXfer:auxv:read+"
+			 ";qXfer:siginfo:read+;qXfer:siginfo:write+"
 			 ";multiprocess+",
 			 dbg->outsize);
 		write_packet(dbg, supported);
 		return 0;
 	}
 	if (!strcmp(name, "Symbol")) {
-		debug("gdb is ready for symbol lookups");
+		LOG(debug) <<"gdb is ready for symbol lookups";
 		dbg->serving_symbol_lookups = 1;
 		write_packet(dbg, "OK");
 		return 0;
@@ -757,7 +801,7 @@ static int query(struct dbg_context* dbg, char* payload)
 		return 1;
 	}
 	if (!strcmp(name, "TStatus")) {
-		debug("gdb asks for trace status");
+		LOG(debug) <<"gdb asks for trace status";
 		/* XXX from the docs, it appears that we should reply
 		 * with "T0" here.  But if we do, gdb keeps bothering
 		 * us with trace queries.  So pretend we don't know
@@ -774,7 +818,7 @@ static int query(struct dbg_context* dbg, char* payload)
 		return xfer(dbg, name, args);
 	}
 
-	unhandled_req(dbg, "Unhandled gdb query: q%s", name);
+	UNHANDLED_REQ(dbg) <<"Unhandled gdb query: q"<< name;
 	return 0;
 }
 
@@ -795,7 +839,7 @@ static int set_var(struct dbg_context* dbg, char* payload)
 		return 0;
 	}
 
-	unhandled_req(dbg, "Unhandled gdb set: Q%s", name);
+	UNHANDLED_REQ(dbg) <<"Unhandled gdb set: Q"<< name;
 	return 0;
 }
 
@@ -824,16 +868,16 @@ static int process_vpacket(struct dbg_context* dbg, char* payload)
 
 		switch (cmd) {
 		case 'C':
-			log_warn("Ignoring request to deliver signal (%s)",
-				 args);
+			LOG(warn) <<"Ignoring request to deliver signal ("
+				  << args <<")";
 			/* fall through */
 		case 'c':
 			dbg->req.type = DREQ_CONTINUE;
 			dbg->req.target = dbg->resume_thread;
 			return 1;
 		case 'S':
-			log_warn("Ignoring request to deliver signal (%s)",
-				 args);
+			LOG(warn) <<"Ignoring request to deliver signal ("
+				  << args <<")";
 			args = strchr(args, ':');
 			if (args) {
 				++args;
@@ -855,14 +899,14 @@ static int process_vpacket(struct dbg_context* dbg, char* payload)
 			}
 			return 1;
 		default:
-			unhandled_req(dbg, "Unhandled vCont command %c(%s)",
-				      cmd, args);
+			UNHANDLED_REQ(dbg) <<"Unhandled vCont command "
+					   << cmd <<"("<< args <<")";
 			return 0;
 		}
 	}
 
 	if (!strcmp("Cont?", name)) {
-		debug("gdb queries which continue commands we support");
+		LOG(debug) <<"gdb queries which continue commands we support";
 		write_packet(dbg, "vCont;c;C;s;S;t;");
 		return 0;
 	}
@@ -872,15 +916,14 @@ static int process_vpacket(struct dbg_context* dbg, char* payload)
 		// assume that this kill request is being made because
 		// a "vRun" restart is coming right up.  We know how
 		// to implement vRun, so we'll ignore this one.
-		debug("gdb asks us to kill tracee(s); ignoring");
+		LOG(debug) <<"gdb asks us to kill tracee(s); ignoring";
 		write_packet(dbg, "OK");
 		return 0;
 	}
 
 	if (!strcmp("Run", name)) {
 		dbg->req.type = DREQ_RESTART;
-		dbg->req.restart.event = -1;
-		dbg->req.restart.port = ntohs(dbg->addr.sin_port);
+		dbg->req.restart.type = RESTART_FROM_PREVIOUS;
 
 		if ('\0' == *args) {
 			return 1;
@@ -888,29 +931,44 @@ static int process_vpacket(struct dbg_context* dbg, char* payload)
 		const char* filename = args;
 		*args++ = '\0';
 		if (strlen(filename)) {
-			fatal("gdb wants us to run the exe image `%s', but we don't support that.",
-				filename);
+			FATAL() <<"gdb wants us to run the exe image `"
+				<< filename << "', but we don't support that.";
 		}
 		if (strchr(args, ';')) {
-			fatal("Extra arguments '%s' passed to run. We don't support that.",
-			      args);
+			FATAL() <<"Extra arguments '"<< args
+				<<"' passed to run. We don't support that.";
 		}
 		if (strlen(args)) {
 			string event_str = decode_ascii_encoded_hex_str(args);
 			char* endp;
-			dbg->req.restart.event =
-				strtol(event_str.c_str(), &endp, 0);
+			// TODO: ideally we would keep checkpointing
+			// out of the gdb protocol translator, and
+			// just pass up the run parameters, but that's
+			// unnecessarily awkward due to the C-style
+			// request struct and the way gdb encodes the
+			// run args.
+			if (event_str[0] == 'c') {
+				int param = strtol(event_str.c_str() + 1, &endp, 0);
+				dbg->req.restart.type = RESTART_FROM_CHECKPOINT;
+				dbg->req.restart.param = param;
+				LOG(debug) <<"next replayer restarting from checkpoint "
+					   << dbg->req.restart.param;
+			} else {
+				dbg->req.restart.type = RESTART_FROM_EVENT;
+				dbg->req.restart.param =
+					strtol(event_str.c_str(), &endp, 0);
+				LOG(debug) <<"next replayer advancing to event "
+					   << dbg->req.restart.param;
+			}
 			if (!endp || *endp != '\0') {
-				fatal("Couldn't parse event string `%s'",
-				      event_str.c_str());
+				FATAL() <<"Couldn't parse event string `"
+					<< event_str << "'";
 			}
 		}
-		debug("next replayer advancing to event %d",
-		      dbg->req.restart.event);
 		return 1;
 	}
 
-	unhandled_req(dbg, "Unhandled gdb vpacket: v%s", name);
+	UNHANDLED_REQ(dbg) <<"Unhandled gdb vpacket: v" << name;
 	return 0;
 }
 
@@ -933,23 +991,23 @@ static int process_packet(struct dbg_context* dbg)
 		dbg->inbuf[dbg->packetend] = '\0';
 	}
 
-	debug("raw request %c(%s)", request, payload);
+	LOG(debug) <<"raw request "<< request <<"("<< payload <<")";
 
 	switch(request) {
 	case INTERRUPT_CHAR:
-		debug("gdb requests interrupt");
+		LOG(debug) <<"gdb requests interrupt";
 		dbg->req.type = DREQ_INTERRUPT;
 		ret = 1;
 		break;
 	case 'D':
-		debug("gdb is detaching from us");
+		LOG(debug) <<"gdb is detaching from us";
 		dbg->req.type = DREQ_DETACH;
 		ret = 1;
 		break;
 	case 'g':
 		dbg->req.type = DREQ_GET_REGS;
 		dbg->req.target = dbg->query_thread;
-		debug("gdb requests registers");
+		LOG(debug) <<"gdb requests registers";
 		ret = 1;
 		break;
 	case 'G':
@@ -969,14 +1027,14 @@ static int process_packet(struct dbg_context* dbg)
 		dbg->req.target = parse_threadid(payload, &payload);
 		assert('\0' == *payload);
 
-		debug("gdb selecting %d.%d",
-		      dbg->req.target.pid, dbg->req.target.tid);
+		LOG(debug) <<"gdb selecting "<< dbg->req.target;
 
 		ret = 1;
 		break;
 	case 'k':
-		log_info("gdb requests kill, exiting");
+		LOG(info) <<"gdb requests kill, exiting";
 		write_packet(dbg, "OK");
+		dbg_destroy_context(&dbg);
 		exit(0);
 	case 'm':
 		dbg->req.type = DREQ_GET_MEM;
@@ -986,32 +1044,39 @@ static int process_packet(struct dbg_context* dbg)
 		dbg->req.mem.len = strtoul(payload, &payload, 16);
 		assert('\0' == *payload);
 
-		debug("gdb requests memory (addr=0x%p, len=%u)",
-			  dbg->req.mem.addr, dbg->req.mem.len);
+		LOG(debug) <<"gdb requests memory (addr="<< dbg->req.mem.addr
+			   <<", len="<< dbg->req.mem.len <<")";
 
 		ret = 1;
 		break;
 	case 'M':
 		/* We can't allow the debugger to write arbitrary data
 		 * to memory, or the replay may diverge. */
+		// TODO: parse this packet in case some oddball gdb
+		// decides to send it instead of 'X'
 		write_packet(dbg, "");
 		ret = 0;
 		break;
 	case 'p':
 		dbg->req.type = DREQ_GET_REG;
 		dbg->req.target = dbg->query_thread;
-		dbg->req.reg = DbgRegister(strtoul(payload, &payload, 16));
+		dbg->req.reg.name = strtoul(payload, &payload, 16);
 		assert('\0' == *payload);
-		debug("gdb requests register value (%d)", dbg->req.reg);
+		LOG(debug) <<"gdb requests register value (" 
+			   << dbg->req.reg.name <<")";
 		ret = 1;
 		break;
 	case 'P':
-		/* XXX we can't let gdb spray registers in general,
-		 * because it may cause replay to diverge.  But some
-		 * writes may be OK.  Let's see how far we can get
-		 * with ignoring these requests. */
-		write_packet(dbg, "");
-		ret = 0;
+		dbg->req.type = DREQ_SET_REG;
+		dbg->req.target = dbg->query_thread;
+		dbg->req.reg.name = strtoul(payload, &payload, 16);
+		assert('=' == *payload++);
+
+		read_reg_value(&payload, &dbg->req.reg);
+
+		assert('\0' == *payload);
+
+		ret = 1;
 		break;
 	case 'q':
 		ret = query(dbg, payload);
@@ -1023,25 +1088,41 @@ static int process_packet(struct dbg_context* dbg)
 		dbg->req.type = DREQ_GET_IS_THREAD_ALIVE;
 		dbg->req.target = parse_threadid(payload, &payload);
 		assert('\0' == *payload);
-		debug("gdb wants to know if %d.%d is alive",
-		      dbg->req.target.pid, dbg->req.target.tid);
+		LOG(debug) <<"gdb wants to know if "<< dbg->req.target
+		<<" is alive";
 		ret = 1;
 		break;
 	case 'v':
 		ret = process_vpacket(dbg, payload);
 		break;
-	case 'X':
-		/* We can't allow the debugger to write arbitrary data
-		 * to memory, or the replay may diverge. */
-		write_packet(dbg, "");
-		ret = 0;
-		break;
+	case 'X': {
+		dbg->req.type = DREQ_SET_MEM;
+		dbg->req.target = dbg->query_thread;
+		dbg->req.mem.addr = (void*)strtoul(payload, &payload, 16);
+		++payload;
+		dbg->req.mem.len = strtoul(payload, &payload, 16);
+		++payload;
+		// This is freed by dbg_reply_set_mem().
+		dbg->req.mem.data = (byte*)malloc(dbg->req.mem.len);
+		// TODO: verify that the length of |payload| is as
+		// expected in the presence of escaped data.  Right
+		// now this call is potential-buffer-overrun-city.
+		read_binary_data((const byte*)payload,
+				 dbg->req.mem.len,
+				 (byte*)dbg->req.mem.data);
+
+		LOG(debug) <<"gdb setting memory (addr="<< dbg->req.mem.addr
+			   <<", len="<< dbg->req.mem.len <<")";
+
+		ret = 1;
+ 		break;
+	}
 	case 'z':
 	case 'Z': {
 		int type = strtol(payload, &payload, 16);
 		assert(',' == *payload++);
 		if (!(0 <= type && type <= 4)) {
-			log_warn("Unknown watch type %d", type);
+			LOG(warn) << "Unknown watch type "<< type;
 			write_packet(dbg, "");
 			ret = 0;
 			break;
@@ -1054,26 +1135,28 @@ static int process_packet(struct dbg_context* dbg)
 		dbg->req.mem.len = strtoul(payload, &payload, 16);
 		assert('\0' == *payload);
 
-		debug("gdb requests %s breakpoint (addr=%p, len=%u)",
-		      'Z' == request ? "set" : "remove",
-		      dbg->req.mem.addr, dbg->req.mem.len);
+		LOG(debug) <<"gdb requests "
+			   << ('Z' == request ? "set" : "remove")
+			   << "breakpoint (addr="<< dbg->req.mem.addr
+			   << ", len=" << dbg->req.mem.len <<")";
 
 		ret = 1;
 		break;
 	}
 	case '!':
-		debug("gdb requests extended mode");
+		LOG(debug) <<"gdb requests extended mode";
 		write_packet(dbg, "OK");
 		ret = 0;
 		break;
 	case '?':
-		debug("gdb requests stop reason");
+		LOG(debug) <<"gdb requests stop reason";
 		dbg->req.type = DREQ_GET_STOP_REASON;
 		dbg->req.target = dbg->query_thread;
 		ret = 1;
 		break;
 	default:
-		unhandled_req(dbg, "Unhandled gdb request '%c'", dbg->inbuf[1]);
+		UNHANDLED_REQ(dbg) <<"Unhandled gdb request '"
+				   << dbg->inbuf[1] <<"'";
 		ret = 0;
 	}
 	/* Erase the newly processed packet from the input buffer. */
@@ -1098,10 +1181,10 @@ void dbg_notify_no_such_thread(struct dbg_context* dbg,
 	 * request has died, and either gdb didn't notice that, or rr
 	 * didn't notify gdb.  Either way, the user should restart
 	 * their debugging session. */
-	log_err(
+	LOG(error) <<
 "Targeted thread no longer exists; this is the result of either a gdb or\n"
 "rr bug.  Please restart your debugging session and avoid doing whatever\n"
-"triggered this bug.");
+"triggered this bug.";
 	write_packet(dbg, "E10");
 	consume_request(dbg);
 }
@@ -1126,7 +1209,7 @@ static void dbg_notify_restart(struct dbg_context* dbg)
 struct dbg_request dbg_get_request(struct dbg_context* dbg)
 {
 	if (DREQ_RESTART == dbg->req.type) {
-		debug("consuming RESTART request");
+		LOG(debug) <<"consuming RESTART request";
 		dbg_notify_restart(dbg);
 		// gdb wants to be notified with a stop packet when
 		// the process "relaunches".  In rr's case, the
@@ -1238,47 +1321,69 @@ static int to_gdb_signum(int sig)
 	case SIGPWR: return 32;
 	case SIGSYS: return 12;
 	default:
-		fatal("Unknown signal %d", sig);
+		FATAL() << "Unknown signal " << sig;
+		return -1;	// not reached
 	}
 }
 
 static void send_stop_reply_packet(struct dbg_context* dbg,
-				   dbg_threadid_t thread, int sig)
+				   dbg_threadid_t thread, int sig,
+				   void* watch_addr = nullptr)
 {
-	if (sig >= 0) {
-		char buf[64];
-		snprintf(buf, sizeof(buf) - 1, "T%02xthread:p%02x.%02x;",
-			 to_gdb_signum(sig), thread.pid, thread.tid);
-		write_packet(dbg, buf);
-	} else {
+	if (sig < 0) {
 		write_packet(dbg, "E01");
+		return;
 	}
+	char watch[1024];
+	if (watch_addr) {
+		snprintf(watch, sizeof(watch) - 1, "watch:%" PRIxPTR ";",
+			 uintptr_t(watch_addr));
+	} else {
+		watch[0] = '\0';
+	}
+	char buf[PATH_MAX];
+	snprintf(buf, sizeof(buf) - 1, "T%02xthread:p%02x.%02x;%s",
+		 to_gdb_signum(sig), thread.pid, thread.tid, watch);
+	write_packet(dbg, buf);
 }
 
-void dbg_notify_stop(struct dbg_context* dbg, dbg_threadid_t thread, int sig)
+void dbg_notify_stop(struct dbg_context* dbg, dbg_threadid_t thread, int sig,
+		     void* watch_addr)
 {
 	assert(dbg_is_resume_request(&dbg->req)
 	       || dbg->req.type == DREQ_INTERRUPT);
 
 	if (dbg->tgid != thread.pid) {
-		debug("ignoring stop of thread %d.%d", thread.pid, thread.tid);
+		LOG(debug) <<"ignoring stop of "<< thread
+			   <<" because we're debugging tgid "<< dbg->tgid;
 		// Re-use the existing continue request to advance to
 		// the next stop we're willing to tell gdb about.
 		return;
 	}
-	send_stop_reply_packet(dbg, thread, sig);
+	send_stop_reply_packet(dbg, thread, sig, watch_addr);
 
 	// This isn't documented in the gdb remote protocol, but if we
 	// don't do this, gdb will sometimes continue to send requests
 	// for the previously-stopped thread when it obviously intends
 	// to making requests about the stopped thread.
-	debug("forcing query/resume thread to %d.%d", thread.pid, thread.tid);
+	LOG(debug) <<"forcing query/resume thread to "<< thread;
 	dbg->query_thread = thread;
 	dbg->resume_thread = thread;
 
 	consume_request(dbg);
 }
 
+void dbg_notify_restart_failed(struct dbg_context* dbg)
+{
+	assert(DREQ_RESTART == dbg->req.type);
+
+	// TODO: it's not known by this author whether gdb knows how
+	// to recover from a failed "run" request.
+	write_packet(dbg, "E01");
+
+	consume_request(dbg);
+}
+
 void dbg_reply_get_current_thread(struct dbg_context* dbg,
 				  dbg_threadid_t thread)
 {
@@ -1320,7 +1425,7 @@ void dbg_reply_get_thread_extra_info(struct dbg_context* dbg,
 {
 	assert(DREQ_GET_THREAD_EXTRA_INFO == dbg->req.type);
 
-	debug("thread extra info: '%s'", info);
+	LOG(debug) <<"thread extra info: '" << info << "'";
 	// XXX docs don't say whether we should send the null
 	// terminator.  See what happens.
 	write_hex_bytes_packet(dbg, (const byte*)info, 1 + strlen(info));
@@ -1353,6 +1458,16 @@ void dbg_reply_get_mem(struct dbg_context* dbg, const byte* mem, size_t len)
 	consume_request(dbg);
 }
 
+void dbg_reply_set_mem(struct dbg_context* dbg, int ok)
+{
+	assert(DREQ_SET_MEM == dbg->req.type);
+
+	write_packet(dbg, ok ? "OK" : "E01");
+	free((byte*)dbg->req.mem.data);
+
+	consume_request(dbg);
+}
+
 void dbg_reply_get_offsets(struct dbg_context* dbg/*, TODO */)
 {
 	assert(DREQ_GET_OFFSETS == dbg->req.type);
@@ -1363,53 +1478,52 @@ void dbg_reply_get_offsets(struct dbg_context* dbg/*, TODO */)
 	consume_request(dbg);
 }
 
-/**
- * Format |value| into |buf| in the manner gdb expects.  |buf| must
- * point at a buffer with at least |1 + 2*sizeof(long)| bytes
- * available.  Exactly that many bytes (including '\0' terminator)
- * will be written by this function.
- */
-static void print_reg(dbg_regvalue_t value, char* buf) {
-	if (value.defined) {
-		/* gdb wants the register value in native endianness, so
-		 * swizzle to big-endian so that printf gives us a
-		 * little-endian string.  (Network order is big-endian.) */
-		long v = htonl(value.value);
-		sprintf(buf, "%08lx", v);
-	} else {
-		strcpy(buf, "xxxxxxxx");
-	}
-}
-
-void dbg_reply_get_reg(struct dbg_context* dbg, dbg_regvalue_t value)
+void dbg_reply_get_reg(struct dbg_context* dbg, const DbgRegister& reg)
 {
-	char buf[32];
+	char buf[2 * DBG_MAX_REG_SIZE + 1];
 
 	assert(DREQ_GET_REG == dbg->req.type);
 
-	print_reg(value, buf);
+	print_reg_value(reg, buf);
 	write_packet(dbg, buf);
 
 	consume_request(dbg);
 }
 
-void dbg_reply_get_regs(struct dbg_context* dbg,
-			const struct dbg_regfile* file)
+void dbg_reply_get_regs(struct dbg_context* dbg, const DbgRegfile& file)
 {
-	/* XXX this will be wrong on x64 WINNT */
-	char buf[1 + DREG_NUM_LINUX_I386 * 2 * sizeof(long)];
-	int i;
+	size_t n_regs = file.total_registers();
+	char buf[n_regs * 2 * DBG_MAX_REG_SIZE + 1];
 
 	assert(DREQ_GET_REGS == dbg->req.type);
 
-	for (i = 0; i < DREG_NUM_LINUX_I386; ++i) {
-		print_reg(file->regs[i], &buf[i * 2 * sizeof(long)]);
+	size_t offset = 0;
+	for (auto it = file.regs.begin(), end = file.regs.end();
+	     it != end; ++it) {
+		offset += print_reg_value(*it, &buf[offset]);
 	}
 	write_packet(dbg, buf);
 
 	consume_request(dbg);
 }
 
+void dbg_reply_set_reg(struct dbg_context* dbg, bool ok)
+{
+	assert(DREQ_SET_REG == dbg->req.type);
+
+	// TODO: what happens if we're forced to reply to a
+	// set-register request with |ok = false|, leading us to
+	// pretend not to understand the packet?  If, later, an
+	// experimental session needs the set-register request will it
+	// not be sent?
+	//
+	// We can't reply with an error packet here because gdb thinks
+	// that failed set-register requests are catastrophic.
+	write_packet(dbg, ok ? "OK" : "");
+
+	consume_request(dbg);
+}
+
 void dbg_reply_get_stop_reason(struct dbg_context* dbg,
 			       dbg_threadid_t which, int sig)
 {
@@ -1458,13 +1572,7 @@ void dbg_reply_watchpoint_request(struct dbg_context* dbg, int code)
 	assert(DREQ_WATCH_FIRST <= dbg->req.type
 	       && dbg->req.type <= DREQ_WATCH_LAST);
 
-	if (code) {
-		fprintf(stderr,
-"rr: Warning: attempt to set unhandled watchpoint type.\n"
-"  Use 'del [breakpoint-num]' to clear the watchpoint, or else gdb may \n"
-"  become unusable.\n");
-	}
-	write_packet(dbg, code ? "" : "OK");
+	write_packet(dbg, code ? "E01" : "OK");
 
 	consume_request(dbg);
 }
@@ -1478,16 +1586,37 @@ void dbg_reply_detach(struct dbg_context* dbg)
 	consume_request(dbg);
 }
 
+void dbg_reply_read_siginfo(struct dbg_context* dbg,
+			    const byte* si_bytes, ssize_t num_bytes)
+{
+	assert(DREQ_READ_SIGINFO == dbg->req.type);
+
+	if (num_bytes < 0) {
+		write_packet(dbg, "E01");
+	} else {
+		write_binary_packet(dbg, "l", si_bytes, num_bytes);
+	}
+
+	consume_request(dbg);
+}
+
+void dbg_reply_write_siginfo(struct dbg_context* dbg/*, TODO*/)
+{
+	assert(DREQ_WRITE_SIGINFO == dbg->req.type);
+
+	write_packet(dbg, "E01");
+
+	consume_request(dbg);
+}
+
 void dbg_destroy_context(struct dbg_context** dbg)
 {
 	struct dbg_context* d;
 	if (!(d = *dbg)) {
 		return;
 	}
-	*dbg = NULL;
-	close(d->desc_fd);
 	close(d->listen_fd);
 	close(d->sock_fd);
-	munmap(d, dbg_context_mapped_size());
+	free(d);
 	*dbg = NULL;
 }
