@@ -3,23 +3,28 @@
 #include <assert.h>
 #include <dlfcn.h>
 #include <getopt.h>
+#include <inttypes.h>
+#include <linux/version.h>
 #include <sched.h>
 #include <stdio.h>
 #include <string.h>
 #include <sysexits.h>
 #include <sys/prctl.h>
+#include <sys/utsname.h>
 #include <sys/wait.h>
 
 #include <limits>
+#include <sstream>
 
 #include "preload/syscall_buffer.h"
 
 #include "config.h"
-#include "dbg.h"
 #include "hpc.h"
+#include "log.h"
 #include "recorder.h"
 #include "recorder_sched.h"
 #include "replayer.h"
+#include "syscalls.h"
 #include "task.h"
 #include "trace.h"
 #include "util.h"
@@ -28,6 +33,40 @@ using namespace std;
 
 extern char** environ;
 
+static void dump_syscallbuf_data(TraceIfstream& trace, FILE* out,
+				 const struct trace_frame& frame)
+{
+	if (frame.ev.type != EV_SYSCALLBUF_FLUSH) {
+		return;
+	}
+	struct raw_data buf;
+	trace >> buf;
+	if (buf.global_time != frame.global_time || buf.ev != frame.ev) {
+		fprintf(stderr, "Malformed trace file (time+event mismatch)\n");
+		abort();
+	}
+	size_t bytes_remaining = buf.data.size() - sizeof(sizeof(struct syscallbuf_hdr));
+	auto flush_hdr = reinterpret_cast<const syscallbuf_hdr*>(buf.data.data());
+	if (flush_hdr->num_rec_bytes != bytes_remaining) {
+		fprintf(stderr, "Malformed trace file (bad recorded-bytes count)\n");
+		abort();
+	}
+
+	auto record_ptr = reinterpret_cast<const uint8_t*>(flush_hdr + 1);
+	auto end_ptr = record_ptr + bytes_remaining;
+	while (record_ptr < end_ptr) {
+		auto record = reinterpret_cast<const struct syscallbuf_record*>(record_ptr);
+		fprintf(out, "  { syscall:'%s', ret:0x%lx }\n",
+			syscallname(record->syscallno, frame.ev.arch()),
+			record->ret);
+		if (record->size < sizeof(*record)) {
+			fprintf(stderr, "Malformed trace file (bad record size)\n");
+			abort();
+		}
+		record_ptr += stored_record_size(record->size);
+	}
+}
+
 /**
  * Dump all events from the current to trace that match |spec| to
  * |out|.  |spec| has the following syntax: /\d+(-\d+)?/, expressing
@@ -40,88 +79,118 @@ extern char** environ;
  * is, they should comprise disjoint and monotonically increasing
  * event sets.  No attempt is made to enforce this or normalize specs.
  */
-static void dump_events_matching(FILE* out, const char* spec)
+static void dump_events_matching(TraceIfstream& trace,
+				 FILE* out, const char* spec)
 {
+	
 	uint32_t start = 0, end = numeric_limits<uint32_t>::max();
-	struct trace_frame frame;
 
-	/* Try to parse the "range" syntax '[start]-[end]'. */
+	// Try to parse the "range" syntax '[start]-[end]'.
 	if (spec && 2 > sscanf(spec, "%u-%u", &start, &end)) {
-		/* Fall back on assuming the spec is a single event
-		 * number, however it parses out with atoi(). */
+		// Fall back on assuming the spec is a single event
+		// number, however it parses out with atoi().
 		start = end = atoi(spec);
 	}
 
-	while (try_read_next_trace(&frame) && frame.global_time <= end) {
-		if (start <= frame.global_time) {
-			dump_trace_frame(out, &frame);
+	bool dump_raw_data = rr_flags()->dump_syscallbuf;
+	while (!trace.at_end()) {
+		struct trace_frame frame;
+		trace >> frame;
+		if (end < frame.global_time) {
+			return;
+		}
+		if (start <= frame.global_time && frame.global_time <= end) {
+			frame.dump(out, rr_flags()->raw_dump);
+			if (rr_flags()->dump_syscallbuf) {
+				dump_syscallbuf_data(trace, out, frame);
+			}
+			if (!rr_flags()->raw_dump) {
+				fprintf(out, "}\n");
+			}
+		}
+		struct raw_data data;
+		while (dump_raw_data && trace.read_raw_data_for_frame(frame, data)) {
+			// Skip raw data for this frame
 		}
 	}
 }
 
-static void start_dumping(int argc, char* argv[], char** envp)
+static void dump_statistics(const TraceIfstream& trace, FILE* out)
+{
+	uint64_t uncompressed = trace.uncompressed_bytes();
+	uint64_t compressed = trace.compressed_bytes();
+	fprintf(stdout, "// Uncompressed bytes %" PRIu64 ", compressed bytes %" PRIu64 ", ratio %.2fx\n",
+		uncompressed, compressed, double(uncompressed)/compressed);
+}
+
+static int dump(int argc, char* argv[], char** envp)
 {
 	FILE* out = stdout;
+	auto trace = TraceIfstream::open(argc, argv);
 
-	rep_setup_trace_dir(argv[0]);
-	open_trace_files();
-	rep_init_trace_files();
-
-	fprintf(out,
-		"global_time thread_time tid reason entry/exit "
-		"hw_interrupts page_faults adapted_rbc instructions "
-		"eax ebx ecx edx esi edi ebp orig_eax esp eip eflags\n");
+	if (rr_flags()->raw_dump) {
+		fprintf(out,
+			"global_time thread_time tid reason "
+			"hw_interrupts page_faults adapted_rbc instructions "
+			"eax ebx ecx edx esi edi ebp orig_eax esp eip eflags\n");
+	}
 
 	if (1 == argc) {
 		// No specs => dump all events.
-		return dump_events_matching(stdout, NULL /*all events*/);
+		dump_events_matching(*trace, stdout,
+				     nullptr/*all events*/);
+	} else {
+		for (int i = 1; i < argc; ++i) {
+			dump_events_matching(*trace, stdout, argv[i]);
+		}
 	}
 
-	for (int i = 1; i < argc; ++i) {
-		dump_events_matching(stdout, argv[i]);
+	if (rr_flags()->dump_statistics) {
+		dump_statistics(*trace, stdout);
 	}
+
+	return 0;
 }
 
-static void start(const char* rr_exe, int argc, char* argv[], char** envp)
+static int start(const char* rr_exe, int argc, char* argv[], char** envp)
 {
-
 	switch (rr_flags()->option) {
 	case RECORD:
 		return record(rr_exe, argc, argv, envp);
 	case REPLAY:
 		return replay(argc, argv, envp);
 	case DUMP_EVENTS:
-		return start_dumping(argc, argv, envp);
+		return dump(argc, argv, envp);
 	default:
-		fatal("Uknown option %d", rr_flags()->option);
+		FATAL() <<"Unknown option "<< rr_flags()->option;
+		return 0; // unreached
 	}
 }
 
-/**
- * Open |filename| and scan it as if it contains a single integer
- * value.  Return the integer on success, or -1 on failure.
- */
-static int read_int_file(const char* filename)
+static void assert_prerequisites(struct flags* flags)
 {
-	FILE* inf = fopen(filename, "r");
-	int val;
-	if (!inf) {
-		return -1;
+	struct utsname uname_buf;
+	memset(&uname_buf, 0, sizeof(uname_buf));
+	if (!uname(&uname_buf)) {
+		unsigned int major, minor;
+		char dot;
+		stringstream stream(uname_buf.release);
+		stream >> major >> dot >> minor;
+		if (KERNEL_VERSION(major, minor, 0) < KERNEL_VERSION(3, 4, 0)) {
+			FATAL() << "Kernel doesn't support necessary ptrace "
+				<< "functionality; need 3.4.0 or better.";
+		}
+
+		if (flags->use_syscall_buffer &&
+		    KERNEL_VERSION(major, minor, 0) < KERNEL_VERSION(3, 5, 0)) {
+			FATAL() << "Your kernel does not support syscall "
+				<< "filtering; please use the -n option";
+		}
 	}
-	if (1 != fscanf(inf, "%d", &val)) {
-		fatal("Failed to scan integer from %s", filename);
-	}
-	fclose(inf);
-	return val;
 }
 
-static void assert_prerequisites(void) {
-	int ptrace_scope_val =
-		read_int_file("/proc/sys/kernel/yama/ptrace_scope");
-	if (ptrace_scope_val > 0) {
-		fatal("Can't write to process memory; ptrace_scope is %d",
-		      ptrace_scope_val);
-	}
+static void check_performance_settings()
+{
 	// NB: we hard-code "cpu0" here because rr pins itself and all
 	// tracees to cpu 0.  We don't care about the other CPUs.
 	int fd = open("/sys/devices/system/cpu/cpu0/cpufreq/scaling_governor",
@@ -130,13 +199,13 @@ static void assert_prerequisites(void) {
 		// If the file doesn't exist, the system probably
 		// doesn't have the ability to frequency-scale, for
 		// example a VM.
-		log_warn("Unable to check CPU-frequency governor.");
+		LOG(info) <<"Unable to check CPU-frequency governor.";
 		return;
 	}
 	char governor[PATH_MAX];
 	ssize_t nread = read(fd, governor, sizeof(governor) - 1);
 	if (0 > nread) {
-		fatal("Unable to read cpu0's frequency governor.");
+		FATAL() <<"Unable to read cpu0's frequency governor.";
 	}
 	governor[nread] = '\0';
 	ssize_t len = strlen(governor);
@@ -144,7 +213,7 @@ static void assert_prerequisites(void) {
 		// Eat the '\n'.
 		governor[len - 1] = '\0';
 	}
-	log_info("cpu0's frequency governor is '%s'", governor);
+	LOG(info) <<"cpu0's frequency governor is '"<< governor <<"'";
 	if (strcmp("performance", governor)) {
 		fprintf(stderr,
 "\n"
@@ -164,6 +233,7 @@ static void assert_prerequisites(void) {
 		// clever to enable the 'performance' just for rr, but
 		// that seems too hard at the moment.
 	}
+	close(fd);
 }
 
 static void print_usage(void)
@@ -172,6 +242,10 @@ static void print_usage(void)
 "Usage: rr [OPTION] (record|replay|dump) [OPTION]... [ARG]...\n"
 "\n"
 "Common options\n"
+"  -a, --microarch=<NAME>     force rr to assume it's running on a CPU\n"
+"                             with microarch NAME even if runtime detection\n"
+"                             says otherwise.  NAME should be a string like\n"
+"                             'Ivy Bridge'.\n"
 "  -c, --checksum={on-syscalls,on-all-events}|FROM_TIME\n"
 "                             compute and store (during recording) or\n"
 "                             read and verify (during replay) checksums\n"
@@ -184,14 +258,19 @@ static void print_usage(void)
 "                             file `[trace_dir]/[tid].[time]_{rec,rep}':\n"
 "                             `_rec' for dumps during recording, `_rep'\n"
 "                             for dumps during replay\n"
-"  -f, --force-enable-debugger\n"
-"                             always allow emergency debugging, even\n"
-"                             when it doesn't seem like a good idea, for\n"
-"                             example if stderr isn't a tty.\n"
+"  -f, --force-things\n       force rr to do some things that don't seem\n"
+"                             like good ideas, for example launching an\n"
+"                             interactive emergency debugger if stderr\n"
+"                             isn't a tty.\n"
 "  -k, --check-cached-mmaps   verify that cached task mmaps match /proc/maps\n"
+"  -e, --fatal-errors         any warning or error that is printed is\n"
+"                             treated as fatal\n"
 "  -m, --mark-stdio           mark stdio writes with [rr.<EVENT-NO>],\n"
 "                             where EVENT-NO is the global trace time at\n"
 "                             which the write occures.\n"
+"  -s, --suppress-environment-warnings\n"
+"                             suppress warnings about issues in the\n"
+"                             environment that rr has no control over\n"
 "  -t, --dump-at=TIME         dump memory at global timepoint TIME\n"
 "  -u, --cpu-unbound          allow tracees to run on any virtual CPU.\n"
 "                             Default is to bind to CPU 0.  This option\n"
@@ -219,7 +298,7 @@ static void print_usage(void)
 "                             even if it would otherwise be used\n"
 "\n"
 "Syntax for `replay'\n"
-" rr replay [OPTION]... <trace-dir>\n"
+" rr replay [OPTION]... [<trace-dir>]\n"
 "  -a, --autopilot            replay without debugger server\n"
 "  -f, --onfork=<PID>         start a debug server when <PID> has been\n"
 "                             fork()d, AND the target event has been\n"
@@ -233,14 +312,17 @@ static void print_usage(void)
 "  -s, --dbgport=<PORT>       only start a debug server on <PORT>;\n"
 "                             don't automatically launch the debugger\n"
 "                             client too.\n"
+"  -x, --gdb-x=<FILE>         execute gdb commands from <FILE>\n"
 "\n"
 "Syntax for `dump`\n"
-" rr dump [OPTIONS] <trace_dir> <event-spec>...\n"
+" rr dump [OPTIONS] <trace_dir> [<event-spec>...]\n"
 "  Event specs can be either an event number like `127', or a range\n"
 "  like `1000-5000'.  By default, all events are dumped.\n"
 "  -r, --raw                  dump trace frames in a more easily\n"
 "                             machine-parseable format instead of the\n"
 "                             default human-readable format\n"
+"  -s, --statistics           dump statistics about the trace\n"
+"  -b, --syscallbuf           dump syscallbuf contents\n"
 "\n"
 "A command line like `rr (-h|--help|help)...' will print this message.\n"
 , stderr);
@@ -257,7 +339,7 @@ static int parse_record_args(int cmdi, int argc, char** argv,
 		{ "no-syscall-buffer", no_argument, NULL, 'n' },
 		{ 0 }
 	};
-	optind = cmdi + 1;
+	optind = cmdi;
 	while (1) {
 		int i = 0;
 		switch (getopt_long(argc, argv, "+c:be:i:n", opts, &i)) {
@@ -295,12 +377,13 @@ static int parse_replay_args(int cmdi, int argc, char** argv,
 		{ "no-redirect-output", no_argument, NULL, 'q' },
 		{ "onfork", required_argument, NULL, 'f' },
 		{ "onprocess", required_argument, NULL, 'p' },
+		{ "gdb-x", required_argument, NULL, 'x' },
 		{ 0 }
 	};
-	optind = cmdi + 1;
+	optind = cmdi;
 	while (1) {
 		int i = 0;
-		switch (getopt_long(argc, argv, "+af:g:p:qs:", opts, &i)) {
+		switch (getopt_long(argc, argv, "+af:g:p:qs:x:", opts, &i)) {
 		case -1:
 			return optind;
 		case 'a':
@@ -326,6 +409,9 @@ static int parse_replay_args(int cmdi, int argc, char** argv,
 			flags->dbgport = atoi(optarg);
 			flags->dont_launch_debugger = true;
 			break;
+		case 'x':
+			flags->gdb_command_file_path = optarg;
+			break;
 		default:
 			return -1;
 		}
@@ -336,17 +422,26 @@ static int parse_dump_args(int cmdi, int argc, char** argv,
 			   struct flags* flags)
 {
 	struct option opts[] = {
+		{ "syscallbuf", no_argument, NULL, 'b' },
 		{ "raw", no_argument, NULL, 'r' },
+		{ "statistics", no_argument, NULL, 's' },
+		{ 0 }
 	};
-	optind = cmdi + 1;
+	optind = cmdi;
 	while (1) {
 		int i = 0;
-		switch (getopt_long(argc, argv, "r", opts, &i)) {
+		switch (getopt_long(argc, argv, "brs", opts, &i)) {
 		case -1:
 			return optind;
+		case 'b':
+			flags->dump_syscallbuf = true;
+			break;
 		case 'r':
 			flags->raw_dump = true;
 			break;
+		case 's':
+			flags->dump_statistics = true;
+			break;
 		default:
 			return -1;
 		}
@@ -361,36 +456,44 @@ static int parse_common_args(int argc, char** argv, struct flags* flags)
 		{ "cpu-unbound", no_argument, NULL, 'u' },
 		{ "dump-at", required_argument, NULL, 't' },
 		{ "dump-on", required_argument, NULL, 'd' },
-		{ "force-enable-debugger", no_argument, NULL, 'f' },
+		{ "force-things", no_argument, NULL, 'f' },
+		{ "force-microarch", required_argument, NULL, 'a' },
 		{ "mark-stdio", no_argument, NULL, 'm' },
+		{ "suppress-environment-warnings", no_argument, NULL, 's' },
+		{ "fatal-errors", no_argument, NULL, 'e' },
 		{ "verbose", no_argument, NULL, 'v' },
 		{ "wait-secs", required_argument, NULL, 'w' },
 		{ 0 }
 	};
 	while (1) {
 		int i = 0;
-		switch (getopt_long(argc, argv, "+c:d:fkmt:uvw:", opts, &i)) {
+		switch (getopt_long(argc, argv, "+a:c:d:fkmst:uvw:", opts, &i)) {
 		case -1:
 			return optind;
+		case 'a':
+			flags->forced_uarch = optarg;
+			break;
 		case 'c':
 			if (!strcmp("on-syscalls", optarg)) {
-				log_info("checksumming on syscall exit");
+				LOG(info) <<"checksumming on syscall exit";
 				flags->checksum = CHECKSUM_SYSCALL;
 			} else if (!strcmp("on-all-events", optarg)) {
-				log_info("checksumming on all events");
+				LOG(info) <<"checksumming on all events";
 				flags->checksum = CHECKSUM_ALL;
 			} else {
-				flags->checksum = str2li(optarg,
-							 LI_COLUMN_SIZE);
-				log_info("checksumming on at event %d",
-					 flags->checksum);
+				flags->checksum = atoi(optarg);
+				LOG(info) <<"checksumming on at event "
+					  << flags->checksum;
 			}
 			break;
 		case 'd':
 			flags->dump_on = atoi(optarg);
 			break;
+		case 'e':
+			flags->fatal_errors_and_warnings = true;
+			break;
 		case 'f':
-			flags->force_enable_debugger = true;
+			flags->force_things = true;
 			break;
 		case 'k':
 			flags->check_cached_mmaps = true;
@@ -398,6 +501,9 @@ static int parse_common_args(int argc, char** argv, struct flags* flags)
 		case 'm':
 			flags->mark_stdio = true;
 			break;
+		case 's':
+			flags->suppress_environment_warnings = true;
+			break;
 		case 't':
 			flags->dump_at = atoi(optarg);
 			break;
@@ -423,7 +529,6 @@ static int parse_args(int argc, char** argv, struct flags* flags)
 	const char* cmd;
 	int cmdi;
 
-	memset(flags, 0, sizeof(*flags));
 	flags->max_rbc = DEFAULT_MAX_RBC;
 	flags->max_events = DEFAULT_MAX_EVENTS;
 	flags->checksum = CHECKSUM_NONE;
@@ -432,6 +537,7 @@ static int parse_args(int argc, char** argv, struct flags* flags)
 	flags->dump_on = DUMP_ON_NONE;
 	flags->redirect = true;
 	flags->use_syscall_buffer = true;
+	flags->suppress_environment_warnings = false;
 
 	if (0 > (cmdi = parse_common_args(argc, argv, flags))) {
 		return -1;
@@ -444,22 +550,48 @@ static int parse_args(int argc, char** argv, struct flags* flags)
 	cmd = argv[cmdi];
 	if (!strcmp("record", cmd)) {
 		flags->option = RECORD;
-		return parse_record_args(cmdi, argc, argv, flags);
+		return parse_record_args(cmdi + 1, argc, argv, flags);
 	}
 	if (!strcmp("replay", cmd)) {
 		flags->option = REPLAY;
-		return parse_replay_args(cmdi, argc, argv, flags);
+		return parse_replay_args(cmdi + 1, argc, argv, flags);
 	}
 	if (!strcmp("dump", cmd)) {
 		flags->option = DUMP_EVENTS;
-		return parse_dump_args(cmdi, argc, argv, flags);
+		return parse_dump_args(cmdi + 1, argc, argv, flags);
 	}
 	if (!strcmp("help", cmd) || !strcmp("-h", cmd)
 	    || !strcmp("--help", cmd)) {
 		return -1;
 	}
-	fprintf(stderr, "%s: unknown command `%s`\n", exe, cmd);
-	return -1;
+	flags->option = RECORD;
+	return parse_record_args(cmdi, argc, argv, flags);
+}
+
+static string find_syscall_buffer_library()
+{
+	char* exe_path = realpath("/proc/self/exe", NULL);
+	string lib_path = exe_path;
+	free(exe_path);
+
+	int end = lib_path.length();
+	// Chop off the filename
+	while (end > 0 && lib_path[end - 1] != '/') {
+		--end;
+	}
+	lib_path.erase(end);
+	lib_path += "../lib/" SYSCALLBUF_LIB_FILENAME;
+	if (access(lib_path.c_str(), F_OK) != 0) {
+		// File does not exist. Assume install put it in LD_LIBRARY_PATH.
+		lib_path = SYSCALLBUF_LIB_FILENAME;
+	}
+	return lib_path;
+}
+
+static void init_random()
+{
+	// Not very good, but good enough for our non-security-sensitive needs.
+	srandom(time(NULL) ^ getpid());
 }
 
 int main(int argc, char* argv[])
@@ -468,6 +600,8 @@ int main(int argc, char* argv[])
 	int wait_secs;
 	struct flags* flags = rr_flags_for_init();
 
+	init_random();
+
 	if (argc >= 2 && !strcmp("check-preload-lib", argv[1])) {
 		// If we reach here and we were checking the preload
 		// lib, then it didn't load --- its __constructor__
@@ -475,52 +609,35 @@ int main(int argc, char* argv[])
 		_exit(EX_CONFIG);
 	}
 
-	assert_prerequisites();
-
-	if (0 > (argi = parse_args(argc, argv, flags)) || argc <= argi) {
+	if (0 > (argi = parse_args(argc, argv, flags))
+	    || argc < argi
+	    // |rr replay| is allowed to have no arguments to replay
+	    // the most recently saved trace.
+	    || (REPLAY != flags->option && argc <= argi)) {
 		print_usage();
 		return 1;
 	}
 
+	assert_prerequisites(flags);
+	if (!flags->suppress_environment_warnings) {
+		check_performance_settings();
+	}
+
 	wait_secs = flags->wait_secs;
 	if (wait_secs > 0) {
 		struct timespec ts;
 		ts.tv_sec = wait_secs;
 		ts.tv_nsec = 0;
-		log_info("Waiting %d seconds before continuing ...",
-			 wait_secs);
+		LOG(info) <<"Waiting "<< wait_secs <<" seconds before continuing ...";
 		if (nanosleep_nointr(&ts)) {
-			fatal("Failed to wait requested duration");
-		}
-		log_info("... continuing.");
-	}
-
-	if (!rr_flags()->cpu_unbound) {
-		cpu_set_t mask;
-		// Pin tracee tasks to logical CPU 0, both in
-		// recording and replay.  Tracees can see which HW
-		// thread they're running on by asking CPUID, and we
-		// don't have a way to emulate it yet.  So if a tracee
-		// happens to be scheduled on a different core in
-		// recording than replay, it can diverge.  (And
-		// indeed, has been observed to diverge in practice,
-		// in glibc.)
-		//
-		// Note that this pins both the tracee processes *and*
-		// the tracer procer.  This ends up being a tidy
-		// performance win in certain circumstances,
-		// presumably due to cheaper context switching and/or
-		// better interaction with CPU frequency scaling.
-		CPU_ZERO(&mask);
-		CPU_SET(0, &mask);
-		if (0 > sched_setaffinity(0, sizeof(mask), &mask)) {
-			fatal("Couldn't bind to CPU 0");
+			FATAL() <<"Failed to wait requested duration";
 		}
+		LOG(info) <<"... continuing.";
 	}
 
 	if (RECORD == flags->option) {
-		log_info("Scheduler using max_events=%d, max_rbc=%d",
-			 flags->max_events, flags->max_rbc);
+		LOG(info) << "Scheduler using max_events="<< flags->max_events
+			  <<", max_rbc="<< flags->max_rbc;
 
 		// The syscallbuf library interposes some critical
 		// external symbols like XShmQueryExtension(), so we
@@ -528,18 +645,11 @@ int main(int argc, char* argv[])
 		if (flags->use_syscall_buffer) {
 			setenv(SYSCALLBUF_ENABLED_ENV_VAR, "1", 1);
 		} else {
-			log_info("Syscall buffer disabled by flag");
+			LOG(info) <<"Syscall buffer disabled by flag";
 			unsetenv(SYSCALLBUF_ENABLED_ENV_VAR);
 		}
-		// We rely on the distribution package or the user to
-		// set up the LD_LIBRARY_PATH properly so that we can
-		// LD_PRELOAD the bare library name.  Trying to do
-		// otherwise is possible, but annoying. */
-		flags->syscall_buffer_lib_path = SYSCALLBUF_LIB_FILENAME;
+		flags->syscall_buffer_lib_path = find_syscall_buffer_library();
 	}
 
-	start(argv[0], argc - argi , argv + argi, environ);
-
-	return 0;
-
+	return start(argv[0], argc - argi , argv + argi, environ);
 }
