@@ -62,6 +62,7 @@
 #include <sys/time.h>
 #include <sys/types.h>
 #include <sys/uio.h>
+#include <time.h>
 #include <unistd.h>
 
 /* NB: don't include any other local headers here. */
@@ -78,21 +79,6 @@
 #endif
 #define syscall you_must_use_traced_syscall
 
-/**
- * Copy |_rhs| to |_lhs|.  If the copy overflows, set errno to
- * EOVERFLOW and return -1.
- *
- * WARNING: this macro affects control flow, use with great care.
- */
-#define COPY_CHECK_OVERFLOW(_lhs, _rhs)					\
-	do {								\
-		_lhs = _rhs;						\
-		if (sizeof(_lhs) != sizeof(_rhs) && (_lhs) != (_rhs)) {	\
-			errno = EOVERFLOW;				\
-			return -1;					\
-		}							\
-	} while(0)
-
 /**
  * Represents syscall params.  Makes it simpler to pass them around,
  * and avoids pushing/popping all the data for calls.
@@ -154,11 +140,15 @@ static __thread int desched_counter_fd;
 
 /* Points at the libc/pthread pthread_create().  We wrap
  * pthread_create, so need to retain this pointer to call out to the
- * libc version. */
+ * libc version. There is no __pthread_create stub to call. There are
+ * some explicitly-versioned stubs but let's not use those. */
 static int (*real_pthread_create)(pthread_t* thread,
 				  const pthread_attr_t* attr,
 				  void* (*start_routine) (void*), void* arg);
 
+static int (*real_pthread_mutex_timedlock)(pthread_mutex_t* mutex,
+					   const struct timespec *abstime);
+
 /**
  * Return a pointer to the buffer header, which happens to occupy the
  * initial bytes in the mapped region.
@@ -186,79 +176,49 @@ static byte* buffer_end(void)
 	return buffer + SYSCALLBUF_BUFFER_SIZE;
 }
 
+#define MEMCPY_UNROLL 4
+#define MEMCPY_WORD uintptr_t
+
 /**
  * Same as libc memcpy(), but usable within syscallbuf transaction
  * critical sections.
  */
-static void* local_memcpy(void* dest, const void* source, size_t n)
+static void local_memcpy(void* dest, const void* source, size_t n)
 {
 	char* dst = dest;
 	const char* src = source;
+	static const size_t block_size = MEMCPY_UNROLL*sizeof(MEMCPY_WORD);
+	char* dst_end = dest + n;
 
-	while (n--) *dst++ = *src++;
-
-	return dest;
+	while (dst < dst_end) {
+		if (!((uintptr_t)dst & (block_size - 1))) {
+			// destination is aligned. Use a fast path.
+			size_t words = ((dst_end - dst)/block_size)*MEMCPY_UNROLL;
+			if (words) {
+				const MEMCPY_WORD* block_src = (const MEMCPY_WORD*)src;
+				MEMCPY_WORD* block_dst = (MEMCPY_WORD*)dst;
+				MEMCPY_WORD* block_dst_end = block_dst + words;
+				while (block_dst < block_dst_end) {
+					*block_dst++ = *block_src++;
+					*block_dst++ = *block_src++;
+					*block_dst++ = *block_src++;
+					*block_dst++ = *block_src++;
+				}
+				src = (const char*)block_src;
+				dst = (char*)block_dst;
+				continue;
+			}
+		}
+		*dst++ = *src++;
+	}
 }
 
 /* The following are wrappers for the syscalls invoked by this library
  * itself.  These syscalls will generate ptrace traps. */
 
-/* NB: this code is copied and pasted for both _traced_raw_syscall()
- * and _untraced_raw_syscall() because it needs to be debuggable, and
- * generating it from a macro would hinder that.  So for now, please
- * keep the two in sync.
- *
- * TODO: generate from shared .S file. */
-long _traced_raw_syscall(int syscallno, long a0, long a1, long a2,
-			 long a3, long a4, long a5);
-__asm__(".text\n\t"
-	".globl _traced_raw_syscall\n\t"
-	".type _traced_raw_syscall, @function\n\t"
-	"_traced_raw_syscall:\n\t" /* syscallno = 4(%esp) */
-	".cfi_startproc\n\t"
-
-	"pushl %ebx\n\t"	/* syscallno = 8(%esp) */
-	".cfi_adjust_cfa_offset 4\n\t"
-	".cfi_rel_offset %ebx, 0\n\t"
-	"pushl %esi\n\t"	/* syscallno = 12(%esp) */
-	".cfi_adjust_cfa_offset 4\n\t"
-	".cfi_rel_offset %esi, 0\n\t"
-	"pushl %edi\n\t"	/* syscallno = 16(%esp) */
-	".cfi_adjust_cfa_offset 4\n\t"
-	".cfi_rel_offset %edi, 0\n\t"
-	"pushl %ebp\n\t"	/* syscallno = 20(%esp) */
-	".cfi_adjust_cfa_offset 4\n\t"
-	".cfi_rel_offset %ebp, 0\n\t"
-
-	"movl 20(%esp), %eax\n\t" /* %eax = syscallno */
-	"movl 24(%esp), %ebx\n\t" /* %ebx = a0 */
-	"movl 28(%esp), %ecx\n\t" /* %ecx = a1 */
-	"movl 32(%esp), %edx\n\t" /* %edx = a2 */
-	"movl 36(%esp), %esi\n\t" /* %esi = a3 */
-	"movl 40(%esp), %edi\n\t" /* %edi = a4 */
-	"movl 44(%esp), %ebp\n\t" /* %ebp = a5 */
-
-	"int $0x80\n\t"		/* syscall() */
-	/* When the tracee is in the traced syscall, its $ip will be
-	 * the value of this label.  We need to be able to recognize
-	 * when the tracees are in traced syscalls. */
-	".L_traced_syscall_entry_point_ip:\n\t"
-
-	"popl %ebp\n\t"
-	".cfi_adjust_cfa_offset -4\n\t"
-	".cfi_restore %ebp\n\t"
-	"popl %edi\n\t"
-	".cfi_adjust_cfa_offset -4\n\t"
-	".cfi_restore %edi\n\t"
-	"popl %esi\n\t"
-	".cfi_adjust_cfa_offset -4\n\t"
-	".cfi_restore %esi\n\t"
-	"popl %ebx\n\t"
-	".cfi_adjust_cfa_offset -4\n\t"
-	".cfi_restore %ebx\n\t"
-	"ret\n\t"
-	".cfi_endproc\n\t"
-	".size _traced_raw_syscall, .-_traced_raw_syscall\n\t");
+extern __attribute__((visibility("hidden"))) long
+_traced_raw_syscall(int syscallno, long a0, long a1, long a2,
+		    long a3, long a4, long a5);
 
 static int update_errno_ret(long ret)
 {
@@ -306,16 +266,7 @@ static long traced_raw_syscall(const struct syscall_info* call)
 				   call->args[4], call->args[5]);
 }
 
-static void* get_traced_syscall_entry_point(void)
-{
-    void *ret;
-    __asm__ __volatile__(
-	    "call _get_traced_syscall_entry_point__pic_helper\n\t"
-	    "_get_traced_syscall_entry_point__pic_helper: pop %0\n\t"
-	    "addl $(.L_traced_syscall_entry_point_ip - _get_traced_syscall_entry_point__pic_helper),%0"
-	    : "=a"(ret));
-    return ret;
-}
+extern __attribute__((visibility("hidden"))) void* get_traced_syscall_entry_point(void);
 
 static int traced_fcntl(int fd, int cmd, ...)
 {
@@ -443,62 +394,9 @@ static void exit_signal_critical_section(const sigset_t* saved_mask)
  *
  * XXX make a nice assembly helper like libc's |syscall()|? */
 
-/* NB: this code is copied and pasted for both _traced_raw_syscall()
- * and _untraced_raw_syscall() because it needs to be debuggable, and
- * generating it from a macro would hinder that.  So for now, please
- * keep the two in sync.
- *
- * TODO: generate from shared .S file. */
-long _untraced_raw_syscall(int syscallno, long a0, long a1, long a2,
-			 long a3, long a4, long a5);
-__asm__(".text\n\t"
-	".globl _untraced_raw_syscall\n\t"
-	".type _untraced_raw_syscall, @function\n\t"
-	"_untraced_raw_syscall:\n\t" /* syscallno = 4(%esp) */
-	".cfi_startproc\n\t"
-
-	"pushl %ebx\n\t"	/* syscallno = 8(%esp) */
-	".cfi_adjust_cfa_offset 4\n\t"
-	".cfi_rel_offset %ebx, 0\n\t"
-	"pushl %esi\n\t"	/* syscallno = 12(%esp) */
-	".cfi_adjust_cfa_offset 4\n\t"
-	".cfi_rel_offset %esi, 0\n\t"
-	"pushl %edi\n\t"	/* syscallno = 16(%esp) */
-	".cfi_adjust_cfa_offset 4\n\t"
-	".cfi_rel_offset %edi, 0\n\t"
-	"pushl %ebp\n\t"	/* syscallno = 20(%esp) */
-	".cfi_adjust_cfa_offset 4\n\t"
-	".cfi_rel_offset %ebp, 0\n\t"
-
-	"movl 20(%esp), %eax\n\t" /* %eax = syscallno */
-	"movl 24(%esp), %ebx\n\t" /* %ebx = a0 */
-	"movl 28(%esp), %ecx\n\t" /* %ecx = a1 */
-	"movl 32(%esp), %edx\n\t" /* %edx = a2 */
-	"movl 36(%esp), %esi\n\t" /* %esi = a3 */
-	"movl 40(%esp), %edi\n\t" /* %edi = a4 */
-	"movl 44(%esp), %ebp\n\t" /* %ebp = a5 */
-
-	"int $0x80\n\t"		/* syscall() */
-	/* When the tracee is in the traced syscall, its $ip will be
-	 * the value of this label.  We need to be able to recognize
-	 * when the tracees are in traced syscalls. */
-	".L_untraced_syscall_entry_point_ip:\n\t"
-
-	"popl %ebp\n\t"
-	".cfi_adjust_cfa_offset -4\n\t"
-	".cfi_restore %ebp\n\t"
-	"popl %edi\n\t"
-	".cfi_adjust_cfa_offset -4\n\t"
-	".cfi_restore %edi\n\t"
-	"popl %esi\n\t"
-	".cfi_adjust_cfa_offset -4\n\t"
-	".cfi_restore %esi\n\t"
-	"popl %ebx\n\t"
-	".cfi_adjust_cfa_offset -4\n\t"
-	".cfi_restore %ebx\n\t"
-	"ret\n\t"
-	".cfi_endproc\n\t"
-	".size _untraced_raw_syscall, .-_untraced_raw_syscall\n\t");
+extern __attribute__((visibility("hidden"))) long
+_untraced_raw_syscall(int syscallno, long a0, long a1, long a2,
+		      long a3, long a4, long a5);
 
 /**
  * Unlike |traced_syscall()|, this helper is implicitly "raw" (returns
@@ -525,16 +423,7 @@ static long untraced_syscall(int syscallno, long a0, long a1, long a2,
 #define untraced_syscall0(no)			\
 	untraced_syscall1(no, 0)
 
-static void* get_untraced_syscall_entry_point(void)
-{
-    void *ret;
-    __asm__ __volatile__(
-	    "call _get_untraced_syscall_entry_point__pic_helper\n\t"
-	    "_get_untraced_syscall_entry_point__pic_helper: pop %0\n\t"
-	    "addl $(.L_untraced_syscall_entry_point_ip - _get_untraced_syscall_entry_point__pic_helper),%0"
-	    : "=a"(ret));
-    return ret;
-}
+extern __attribute__((visibility("hidden"))) void* get_untraced_syscall_entry_point(void);
 
 /**
  * Make the *un*traced socketcall |call| with the given args.
@@ -562,116 +451,7 @@ static long untraced_socketcall(int call,
 #define untraced_socketcall0(no)		\
 	untraced_socketcall1(no, 0)
 
-/**
- * Call this hook from |__kernel_vsyscall()|, to buffer syscalls that
- * we otherwise couldn't wrap through LD_PRELOAD helpers.  Return the
- * *RAW* kernel return value, not the -1/errno mandated by POSIX.
- *
- * Remember, this function runs *below* the level of libc.  libc can't
- * know that its call to |__kernel_vsyscall()| has been re-routed to
- * us.
- */
-static long vsyscall_hook(const struct syscall_info* call);
-
-/**
- * |__kernel_vsyscall()| is /actually/ patched to jump here.  This
- * trampoline then prepares a "real" call to |vsyscall_hook()|.
- */
-__asm__(".text\n\t"
-	".globl _vsyscall_hook_trampoline\n\t"
-	".type _vsyscall_hook_trampoline, @function\n\t"
-	"_vsyscall_hook_trampoline:\n\t"
-	".cfi_startproc\n\t"
-
-	/* The monkeypatch pushed $eax on the stack, but there's no
-	 * CFI info for it.  Fix up the CFA offset here to account for
-	 * the monkeypatch code. */
-        ".cfi_adjust_cfa_offset 4\n\t"
-        ".cfi_rel_offset %eax, 0\n\t"
-
-        /* Pull $eax back off the stack.  Now our syscall-arg
-         * registers are restored to their state on entry to
-         * __kernel_vsyscall(). */
-        "popl %eax\n\t"
-	".cfi_adjust_cfa_offset -4\n\t"
-	".cfi_restore %eax\n\t"
-
-        /* Build a |struct syscall_info| by pushing all the syscall
-         * args and the number onto the stack. */
-                                /* struct syscall_info info; */
-        "pushl %ebp\n\t"        /* info.args[5] = $ebp; */
-	".cfi_adjust_cfa_offset 4\n\t"
-	".cfi_rel_offset %ebp, 0\n\t"
-        "pushl %edi\n\t"        /* info.args[4] = $edi; */
-	".cfi_adjust_cfa_offset 4\n\t"
-	".cfi_rel_offset %edi, 0\n\t"
-        "pushl %esi\n\t"        /* info.args[3] = $esi; */
-	".cfi_adjust_cfa_offset 4\n\t"
-	".cfi_rel_offset %esi, 0\n\t"
-        "pushl %edx\n\t"        /* info.args[2] = $edx; */
-	".cfi_adjust_cfa_offset 4\n\t"
-	".cfi_rel_offset %edx, 0\n\t"
-        "pushl %ecx\n\t"        /* info.args[1] = $ecx; */
-	".cfi_adjust_cfa_offset 4\n\t"
-	".cfi_rel_offset %ecx, 0\n\t"
-        "pushl %ebx\n\t"        /* info.args[0] = $ebx; */
-	".cfi_adjust_cfa_offset 4\n\t"
-	".cfi_rel_offset %ebx, 0\n\t"
-        "pushl %eax\n\t"        /* info.no = $eax; */
-	".cfi_adjust_cfa_offset 4\n\t"
-
-        /* $esp points at &info.  Push that pointer on the stack as
-         * our arg for syscall_hook(). */
-        "movl %esp, %ecx\n\t"
-        "pushl %ecx\n\t"
-	".cfi_adjust_cfa_offset 4\n\t"
-
-        "movl $vsyscall_hook, %eax\n\t"
-        "call *%eax\n\t"        /* $eax = vsyscall_hook(&info); */
-
-        /* $eax is now the syscall return value.  Erase the |&info|
-         * arg and |info.no| from the stack so that we can restore the
-         * other registers we saved. */
-        "addl $8, %esp\n\t"
-	".cfi_adjust_cfa_offset -8\n\t"
-
-        /* Contract of __kernel_vsyscall() is that even callee-save
-         * registers aren't touched, so we restore everything here. */
-        "popl %ebx\n\t"
-	".cfi_adjust_cfa_offset -4\n\t"
-	".cfi_restore %ebx\n\t"
-        "popl %ecx\n\t"
-	".cfi_adjust_cfa_offset -4\n\t"
-	".cfi_restore %ecx\n\t"
-        "popl %edx\n\t"
-	".cfi_adjust_cfa_offset -4\n\t"
-	".cfi_restore %edx\n\t"
-        "popl %esi\n\t"
-	".cfi_adjust_cfa_offset -4\n\t"
-	".cfi_restore %esi\n\t"
-        "popl %edi\n\t"
-	".cfi_adjust_cfa_offset -4\n\t"
-	".cfi_restore %edi\n\t"
-        "popl %ebp\n\t"
-	".cfi_adjust_cfa_offset -4\n\t"
-	".cfi_restore %ebp\n\t"
-
-        /* Return to the caller of *|__kernel_vsyscall()|*, because
-         * the monkeypatch jumped to us. */
-        "ret\n\t"
-    	".cfi_endproc\n\t"
-        ".size _vsyscall_hook_trampoline, .-_vsyscall_hook_trampoline\n\t");
-
-static void* get_vsyscall_hook_trampoline(void)
-{
-    void *ret;
-    __asm__ __volatile__(
-	    "call .L_get_vsyscall_hook_trampoline__pic_helper\n\t"
-	    ".L_get_vsyscall_hook_trampoline__pic_helper: pop %0\n\t"
-	    "addl $(_vsyscall_hook_trampoline - .L_get_vsyscall_hook_trampoline__pic_helper),%0"
-	    : "=a"(ret));
-    return ret;
-}
+extern __attribute__((visibility("hidden"))) void _vsyscall_hook_trampoline(void);
 
 /**
  * Do what's necessary to set up buffers for the caller.
@@ -747,7 +527,7 @@ static void install_syscall_filter(void)
  * Return a counter that generates a signal targeted at this task
  * every time the task is descheduled |nr_descheds| times.
  */
-static int open_desched_event_counter(size_t nr_descheds)
+static int open_desched_event_counter(size_t nr_descheds, pid_t tid)
 {
 	struct perf_event_attr attr;
 	int fd;
@@ -768,7 +548,7 @@ static int open_desched_event_counter(size_t nr_descheds)
 		fatal("Failed to fcntl(O_ASYNC) the desched counter");
 	}
 	own.type = F_OWNER_TID;
-	own.pid = traced_gettid();
+	own.pid = tid;
 	if (traced_fcntl(fd, F_SETOWN_EX, &own)) {
 		fatal("Failed to fcntl(SETOWN_EX) the desched counter to this");
 	}
@@ -792,19 +572,20 @@ static void set_up_buffer(void)
 	char cmsgbuf[CMSG_SPACE(sizeof(*cmsg_fdptr))];
 	struct socketcall_args args_vec;
 	struct rrcall_init_buffers_params args;
+	pid_t tid = traced_gettid();
 
 	assert(!buffer);
 
 	/* NB: we want this setup emulated during replay. */
 	if (buffer_enabled) {
-		desched_counter_fd = open_desched_event_counter(1);
+		desched_counter_fd = open_desched_event_counter(1, tid);
 	}
 
 	/* Prepare arguments for rrcall.  We do this in the tracee
 	 * just to avoid some hairy IPC to set up the arguments
 	 * remotely from the tracer; this isn't strictly
 	 * necessary. */
-	prepare_syscallbuf_socket_addr(&addr, traced_gettid());
+	prepare_syscallbuf_socket_addr(&addr, tid);
 
 	memset(&msg, 0, sizeof(msg));
 	msg_fdptr = &msgbuf;
@@ -898,6 +679,8 @@ init_process(void)
 	}
 
 	real_pthread_create = dlsym(RTLD_NEXT, "pthread_create");
+	real_pthread_mutex_timedlock = dlsym(RTLD_NEXT, "pthread_mutex_timedlock");
+
 	buffer_enabled = !!getenv(SYSCALLBUF_ENABLED_ENV_VAR);
 	if (!buffer_enabled) {
 		debug("Syscall buffering is disabled");
@@ -908,7 +691,7 @@ init_process(void)
 	pthread_atfork(NULL, NULL, post_fork_child);
 
 	install_syscall_filter();
-	rrcall_monkeypatch_vdso(get_vsyscall_hook_trampoline());
+	rrcall_monkeypatch_vdso(&_vsyscall_hook_trampoline);
 	process_inited = 1;
 
 	init_thread();
@@ -964,6 +747,65 @@ int pthread_create(pthread_t* thread, const pthread_attr_t* attr,
 	return ret;
 }
 
+#define PTHREAD_MUTEX_TYPE_MASK 3
+#define PTHREAD_MUTEX_PRIO_INHERIT_NP 32
+#define PTHREAD_MUTEX_ELISION_NP 256
+#define PTHREAD_MUTEX_NO_ELISION_NP 512
+
+static void fix_mutex_kind(pthread_mutex_t* mutex)
+{
+	int type = mutex->__data.__kind & PTHREAD_MUTEX_TYPE_MASK;
+	/* Disable priority inheritance. */
+	mutex->__data.__kind &= ~PTHREAD_MUTEX_PRIO_INHERIT_NP;
+	if (type != PTHREAD_MUTEX_TIMED_NP) {
+		/* Currently glibc only tries to use elision for TIMED/NORMAL
+		 * mutexes. Setting elision flag bits for other types of
+		 * mutexes triggers glibc misbehavior.
+		 */
+		return;
+	}
+	/* Cancel explicitly-set elision requests */
+	mutex->__data.__kind &= ~PTHREAD_MUTEX_ELISION_NP;
+	/* Prevent auto-enabling of elision */
+	mutex->__data.__kind |= PTHREAD_MUTEX_NO_ELISION_NP;
+}
+
+/*
+ * We bind directly to __pthread_mutex_lock and __pthread_mutex_trylock
+ * because setting up indirect function pointers in init_process requires
+ * calls to dlsym which itself can call pthread_mutex_lock (e.g. via
+ * application code overriding malloc/calloc to use a pthreads-based
+ * implementation).
+ */
+extern int __pthread_mutex_lock(pthread_mutex_t* mutex);
+extern int __pthread_mutex_trylock(pthread_mutex_t* mutex);
+
+/* Prevent use of lock elision; Haswell's TSX/RTM features used by
+   lock elision increment the rbc perf counter for instructions which
+   are later rolled back if the transaction fails. */
+int pthread_mutex_lock(pthread_mutex_t* mutex)
+{
+	fix_mutex_kind(mutex);
+	return __pthread_mutex_lock(mutex);
+}
+
+int pthread_mutex_timedlock(pthread_mutex_t* mutex,
+			    const struct timespec *abstime)
+{
+	fix_mutex_kind(mutex);
+	/* No __pthread_mutex_timedlock stub exists, so we have to use the
+	 * indirect call.
+	 */
+	return real_pthread_mutex_timedlock(mutex, abstime);
+
+}
+
+int pthread_mutex_trylock(pthread_mutex_t* mutex)
+{
+	fix_mutex_kind(mutex);
+	return __pthread_mutex_trylock(mutex);
+}
+
 /**
  * vsyscall hooks start here.
  *
@@ -1632,7 +1474,7 @@ static long sys_poll(const struct syscall_info* call)
 		 * special case here. */
 		local_memcpy(fds, fds2, nfds * sizeof(*fds));
 	}
-	return commit_raw_syscall(syscallno, ptr, ret);	
+	return commit_raw_syscall(syscallno, ptr, ret);
 }
 
 static long sys_read(const struct syscall_info* call)
@@ -1830,7 +1672,10 @@ static long sys_writev(const struct syscall_info* call)
 	return commit_raw_syscall(syscallno, ptr, ret);
 }
 
-static long __attribute__((unused))
+/* Explicitly declare this as hidden so we can call it from
+ * _vsyscall_hook_trampoline without doing all sorts of special PIC handling.
+ */
+__attribute__((visibility("hidden"))) long
 vsyscall_hook(const struct syscall_info* call)
 {
 	switch (call->no) {
@@ -1893,3 +1738,17 @@ int XShmQueryExtension(void* dpy)
 {
 	return 0;
 }
+
+/** Make sure XShmCreateImage returns null in case an application doesn't do
+    extension checks first. */
+void *XShmCreateImage(register void *dpy,
+	register void *visual,
+	unsigned int depth,
+	int format,
+	char *data,
+	void *shminfo,
+	unsigned int width,
+	unsigned int height)
+{
+	return 0;
+}
