@@ -1,4 +1,4 @@
-/* -*- Mode: C++; tab-width: 8; c-basic-offset: 8; indent-tabs-mode: t; -*- */
+/* -*- mode: C++; tab-width: 8; c-basic-offset: 8; indent-tabs-mode: t; -*- */
 
 //#define DEBUGTAG "Signal"
 
@@ -6,18 +6,21 @@
 
 #include <assert.h>
 #include <fcntl.h>
+#include <linux/perf_event.h>
 #include <sched.h>
 #include <stdlib.h>
 #include <string.h>
 #include <syscall.h>
 #include <sys/mman.h>
 #include <sys/user.h>
+#include <x86intrin.h>
 
 #include "preload/syscall_buffer.h"
 
-#include "dbg.h"
 #include "hpc.h"
+#include "log.h"
 #include "recorder.h"
+#include "remote_syscalls.h"
 #include "task.h"
 #include "trace.h"
 #include "util.h"
@@ -26,9 +29,7 @@ static void handle_siginfo(Task* t, siginfo_t* si);
 
 static __inline__ unsigned long long rdtsc(void)
 {
-	unsigned hi, lo;
-	__asm__ __volatile__ ("rdtsc" : "=a"(lo), "=d"(hi));
-	return ((unsigned long long) lo) | (((unsigned long long) hi) << 32);
+	return __rdtsc();
 }
 
 /**
@@ -43,11 +44,11 @@ static void assert_is_time_slice_interrupt(Task* t, const siginfo_t* si)
 	 * target, because this signal may have become pending before
 	 * we reset the HPC counters.  There be a way to handle that
 	 * more elegantly, but bridge will be crossed in due time. */
-	assert_exec(t, (HPC_TIME_SLICE_SIGNAL == si->si_signo
-			&& si->si_fd == t->hpc->rbc.fd
-			&& POLL_IN == si->si_code),
-		    "Tracee is using SIGSTKFLT??? (code=%d, fd=%d)",
-		    si->si_code, si->si_fd);
+	ASSERT(t, (HPC_TIME_SLICE_SIGNAL == si->si_signo
+		   && si->si_fd == rcb_cntr_fd(t->hpc)
+		   && POLL_IN == si->si_code))
+		<< "Tracee is using SIGSTKFLT??? (code="<< si->si_code
+		<<", fd="<< si->si_fd <<")";
 }
 
 /**
@@ -57,28 +58,32 @@ static void assert_is_time_slice_interrupt(Task* t, const siginfo_t* si)
 static void restore_sigsegv_state(Task* t)
 {
 	kernel_sigaction sa = t->signal_action(SIGSEGV);
-	struct current_state_buffer state;
-	prepare_remote_syscalls(t, &state);
+	AutoRemoteSyscalls remote(t);
 	{
-		struct restore_mem restore;
-		void* child_sa = push_tmp_mem(t, &state, (const byte*)&sa,
-					      sizeof(sa), &restore);
-
-		int ret = remote_syscall4(t, &state, SYS_rt_sigaction,
-					  SIGSEGV, child_sa, NULL, 
-					  _NSIG / 8);
-		assert_exec(t, 0 == ret, "Failed to restore SIGSEGV handler");
-
-		pop_tmp_mem(t, &state, &restore);
+		AutoRestoreMem child_sa(remote, (const byte*)&sa, sizeof(sa));
+		int ret = remote.syscall(SYS_rt_sigaction,
+					 SIGSEGV, static_cast<void*>(child_sa),
+					 nullptr, _NSIG / 8);
+		ASSERT(t, 0 == ret) <<"Failed to restore SIGSEGV handler";
 	}
 	// NB: we would normally want to restore the SIG_BLOCK for
 	// SIGSEGV here, but doing so doesn't change the kernel's
 	// "SigBlk" mask.  There's no bug observed in the kernel's
 	// delivery of SIGSEGV after the RDTSC trap, so we do nothing
 	// here and move on.
-	finish_remote_syscalls(t, &state);
 }
 
+/** Return true iff |t->ip()| points at a RDTSC instruction. */
+static const byte rdtsc_insn[] = { 0x0f, 0x31 };
+static bool is_ip_rdtsc(Task* t)
+{
+	byte insn[sizeof(rdtsc_insn)];
+	if (sizeof(insn) != t->read_bytes_fallible(t->ip(),
+						   sizeof(insn), insn)) {
+		return false;
+	}
+	return !memcmp(insn, rdtsc_insn, sizeof(insn));
+}
 
 /**
  * Return nonzero if |t| was stopped because of a SIGSEGV resulting
@@ -86,78 +91,56 @@ static void restore_sigsegv_state(Task* t)
  */
 static int try_handle_rdtsc(Task *t)
 {
-	int handled = 0;
 	int sig = t->pending_sig();
 	assert(sig != SIGTRAP);
 
-	if (sig <= 0 || sig != SIGSEGV) {
+	if (sig <= 0 || sig != SIGSEGV || !is_ip_rdtsc(t)) {
 		return 0;
 	}
 
-	int size;
-	char *inst = get_inst(t, 0, &size);
-	if (!inst) {
-		/* If the segfault was caused by a jump to a bad $ip,
-		 * then we obviously won't be able to read the
-		 * instruction. */
-		return 0;
+	unsigned long long current_time = rdtsc();
+	Registers r = t->regs();
+	r.set_rdtsc_output(current_time);
+	r.set_ip(r.ip() + sizeof(rdtsc_insn));
+	t->set_regs(r);
+
+	// When SIGSEGV is blocked, apparently the kernel has to do
+	// some ninjutsu to raise the RDTSC trap.  We see the SIGSEGV
+	// bit in the "SigBlk" mask in /proc/status cleared, and if
+	// there's a user handler the SIGSEGV bit in "SigCgt" is
+	// cleared too.  That's perfectly fine, except that it's
+	// unclear who's supposed to undo the signal-state munging.  A
+	// legitimate argument can be made that the tracer is
+	// responsible, so we go ahead and restore the old state.
+	//
+	// One could also argue that this is a kernel bug.  If so,
+	// then this is a workaround that can be removed in the
+	// future.
+	//
+	// If we don't restore the old state, at least firefox has
+	// been observed to hang at delivery of SIGSEGV.  However, the
+	// test written for this bug, fault_in_code_addr, doesn't hang
+	// without the restore.
+	if (t->is_sig_blocked(SIGSEGV)) {
+		restore_sigsegv_state(t);
 	}
 
-	/* if the current instruction is a rdtsc, the segfault was triggered by
-	 * by reading the rdtsc instruction */
-	if (strncmp(inst, "rdtsc", 5) == 0) {
-		long int eax, edx;
-		unsigned long long current_time;
+	t->push_event(Event(EV_SEGV_RDTSC, HAS_EXEC_INFO));
+	LOG(debug) <<"  trapped for rdtsc: returning "<< current_time;
+	return 1;
+}
 
-		current_time = rdtsc();
-		eax = current_time & 0xffffffff;
-		edx = current_time >> 32;
-
-		struct user_regs_struct r = t->regs();
-		r.eax = eax;
-		r.edx = edx;
-		r.eip += size;
-		t->set_regs(r);
-
-		// When SIGSEGV is blocked, apparently the kernel has
-		// to do some ninjutsu to raise the RDTSC trap.  We
-		// see the SIGSEGV bit in the "SigBlk" mask in
-		// /proc/status cleared, and if there's a user handler
-		// the SIGSEGV bit in "SigCgt" is cleared too.  That's
-		// perfectly fine, except that it's unclear who's
-		// supposed to undo the signal-state munging.  A
-		// legitimate argument can be made that the tracer is
-		// responsible, so we go ahead and restore the old
-		// state.
-		//
-		// One could also argue that this is a kernel bug.  If
-		// so, then this is a workaround that can be removed
-		// in the future.
-		//
-		// If we don't restore the old state, at least firefox
-		// has been observed to hang at delivery of SIGSEGV.
-		// However, the test written for this bug,
-		// fault_in_code_addr, doesn't hang without the
-		// restore.
-		if (t->is_sig_blocked(SIGSEGV)) {
-			restore_sigsegv_state(t);
-		}
-
-		t->push_event(Event(EV_SEGV_RDTSC, HAS_EXEC_INFO));
-		handled = 1;
-
-		debug("  trapped for rdtsc: returning %llu", current_time);
+static void arm_desched_event(Task* t)
+{
+	if (ioctl(t->desched_fd, PERF_EVENT_IOC_ENABLE, 0)) {
+		FATAL() <<"Failed to arm desched event";
 	}
-
-	free(inst);
-
-	return handled;
 }
 
 static void disarm_desched_event(Task* t)
 {
 	if (ioctl(t->desched_fd, PERF_EVENT_IOC_DISABLE, 0)) {
-		fatal("Failed to disarm desched event");
+		FATAL() <<"Failed to disarm desched event";
 	}
 }
 
@@ -168,11 +151,11 @@ static void disarm_desched_event(Task* t)
  */
 static void handle_desched_event(Task* t, const siginfo_t* si)
 {
-	assert_exec(t, (SYSCALLBUF_DESCHED_SIGNAL == si->si_signo
-			&& si->si_code == POLL_IN
-			&& si->si_fd == t->desched_fd_child),
-		    "Tracee is using SIGSYS??? (code=%d, fd=%d)",
-		    si->si_code, si->si_fd);
+	ASSERT(t, (SYSCALLBUF_DESCHED_SIGNAL == si->si_signo
+		   && si->si_code == POLL_IN
+		   && si->si_fd == t->desched_fd_child))
+		<<"Tracee is using SIGSYS??? (code="<< si->si_code
+		<<", fd="<< si->si_fd <<")";
 
 	/* If the tracee isn't in the critical section where a desched
 	 * event is relevant, we can ignore it.  See the long comments
@@ -182,7 +165,7 @@ static void handle_desched_event(Task* t, const siginfo_t* si)
 	 * may-block syscall B, but this signal was delivered by an
 	 * event programmed by a previous may-block syscall A. */
 	if (!t->syscallbuf_hdr->desched_signal_may_be_relevant) {
-		debug("  (not entering may-block syscall; resuming)");
+		LOG(debug) <<"  (not entering may-block syscall; resuming)";
 		/* We have to disarm the event just in case the tracee
 		 * has cleared the relevancy flag, but not yet
 		 * disarmed the event itself. */
@@ -286,16 +269,16 @@ static void handle_desched_event(Task* t, const siginfo_t* si)
 		if (SYSCALLBUF_DESCHED_SIGNAL == sig
 		    || HPC_TIME_SLICE_SIGNAL == sig
 		    || t->is_sig_ignored(sig)) {
-			debug("  dropping ignored %s ...", signalname(sig));
+			LOG(debug) <<"  dropping ignored "<< signalname(sig);
 			continue;
 		}
 
-		debug("  stashing %s ...", signalname(sig));
+		LOG(debug) <<"  stashing "<< signalname(sig);
 		t->stash_sig();
 	}
 
 	if (t->is_disarm_desched_event_syscall()) {
-		debug("  (at disarm-desched, so finished buffered syscall; resuming)");
+		LOG(debug) <<"  (at disarm-desched, so finished buffered syscall; resuming)";
 		t->push_event(Event::noop());
 		return;
 	}
@@ -318,7 +301,7 @@ static void handle_desched_event(Task* t, const siginfo_t* si)
 	 * the desched counter when it's trying to step to the entry
 	 * of |call|.  We'll record the syscall entry when the main
 	 * recorder code sees the tracee's syscall event. */
-	record_event(t);
+	t->record_current_event();
 
 	/* Because we set the |delay_syscallbuf_reset| flag and the
 	 * record counter will stay intact for a bit, we need to also
@@ -331,14 +314,21 @@ static void handle_desched_event(Task* t, const siginfo_t* si)
 	 * this one has already been restarted (which we'll detect
 	 * back in the main loop). */
 	t->push_event(Event(interrupted, SyscallEvent(call)));
-	t->ev().Syscall().desched_rec = desched_rec;
-	t->ev().Syscall().regs = t->regs();
-	t->ev().Syscall().state = EXITING_SYSCALL;
+	SyscallEvent& ev = t->ev().Syscall();
+	ev.desched_rec = desched_rec;
+	ev.regs = t->regs();
+	/* For some syscalls (at least poll) but not all (at least not read),
+	 * repeated cont_syscall()s above of the same interrupted syscall
+	 * can set $orig_eax to 0 ... for unclear reasons. Fix that up here
+	 * otherwise we'll get a divergence during replay, which will not
+	 * encounter this problem.
+	 */
+	ev.regs.set_original_syscallno(call);
+	t->set_regs(ev.regs);
+	ev.state = EXITING_SYSCALL;
 
-	debug("  resuming (and probably switching out) blocked `%s'",
-	      syscallname(call));
-
-	return;
+	LOG(debug) <<"  resuming (and probably switching out) blocked `"
+		   << t->syscallname(call) <<"'";
 }
 
 static int is_deterministic_signal(const siginfo_t* si)
@@ -371,10 +361,15 @@ static int is_deterministic_signal(const siginfo_t* si)
 
 static void record_signal(Task* t, const siginfo_t* si)
 {
+	// goto_a_happy_place's stepping can lead to the kernel forgetting
+	// the siginfo for the signal we're going to deliver. Restore that
+	// siginfo now.
+	t->set_siginfo(*si);
+
 	int sig = si->si_signo;
 	if (sig == rr_flags()->ignore_sig) {
-		log_info("Declining to deliver %s by user request",
-			 signalname(sig));
+		LOG(info) <<"Declining to deliver "<< signalname(sig)
+			  <<" by user request";
 		t->push_event(Event::noop());
 		return;
 	}
@@ -445,20 +440,20 @@ static int go_to_a_happy_place(Task* t, siginfo_t* si)
 	 * per above, and if not, steps it until it finds one. */
 	struct syscallbuf_hdr initial_hdr;
 	struct syscallbuf_hdr* hdr = t->syscallbuf_hdr;
+	bool desched_event_maybe_armed = true;
 
-	debug("Stepping tracee to happy place to deliver signal ...");
+	LOG(debug) <<"Stepping tracee to happy place to deliver signal ...";
 
 	if (!hdr) {
 		/* Can't be in critical section because the lock
 		 * doesn't exist yet! */
-		debug("  tracee hasn't allocated syscallbuf yet");
+		LOG(debug) <<"  tracee hasn't allocated syscallbuf yet";
 		goto happy_place;
 	}
 
-	assert_exec(t, !(t->is_in_syscallbuf()
-			 && is_deterministic_signal(si)),
-		    "TODO: %s (code:%d) raised by syscallbuf code",
-		    signalname(si->si_signo), si->si_code);
+	ASSERT(t, !(t->is_in_syscallbuf() && is_deterministic_signal(si)))
+		<< "TODO: "<< signalname(si->si_signo) <<" (code:"
+		<< si->si_code <<") raised by syscallbuf code";
 	/* TODO: when we add support for deterministic signals, we
 	 * should sigprocmask-off all tracee signals while we're
 	 * stepping.  If we tried that with the current impl, the
@@ -477,7 +472,7 @@ static int go_to_a_happy_place(Task* t, siginfo_t* si)
 			 * exception is signal handlers "re-entering"
 			 * desched'd syscalls, which are OK per
 			 * above.. */
-			debug("  tracee outside syscallbuf lib");
+			LOG(debug) <<"  tracee outside syscallbuf lib";
 			goto happy_place;
 		}
 		if (t->is_entering_traced_syscall()) {
@@ -489,95 +484,106 @@ static int go_to_a_happy_place(Task* t, siginfo_t* si)
 			// the syscall entry looks like a synchronous
 			// SIGTRAP generated from the syscallbuf lib,
 			// which we don't know how to handle.
-			debug("  tracee entering traced syscallbuf syscall");
+			LOG(debug) <<"  tracee entering traced syscallbuf syscall";
 			goto happy_place;
 		}
 		if (t->is_traced_syscall()) {
-			debug("  tracee at traced syscallbuf syscall");
+			LOG(debug) <<"  tracee at traced syscallbuf syscall";
 			goto happy_place;
 		}
-		if (t->is_untraced_syscall()
-		    && t->desched_rec()) {
-			debug("  tracee interrupted by desched of %s",
-			      syscallname(t->desched_rec()->syscallno));
+		if (t->is_untraced_syscall() && t->desched_rec()) {
+			LOG(debug) <<"  tracee interrupted by desched of "
+				<< t->syscallname(t->desched_rec()->syscallno);
 			goto happy_place;
 		}
 		if (initial_hdr.locked && !hdr->locked) {
 			/* Tracee just stepped out of a critical
 			 * section and into a happy place.. */
-			debug("  tracee just unlocked syscallbuf");
+			LOG(debug) <<"  tracee just unlocked syscallbuf";
 			goto happy_place;
 		}
 
+		bool should_arm_desched_event = t->is_untraced_syscall();
+		if (desched_event_maybe_armed != should_arm_desched_event) {
+			if (should_arm_desched_event) {
+				arm_desched_event(t);
+			} else {
+				disarm_desched_event(t);
+			}
+			desched_event_maybe_armed = should_arm_desched_event;
+		}
+
 		/* Move the tracee closer to a happy place.  NB: an
 		 * invariant of the syscallbuf is that all untraced
 		 * syscalls must be made from within a transaction
 		 * (critical section), so there's no chance here of
 		 * "skipping over" a syscall we should have
 		 * recorded. */
-		debug("  stepi out of syscallbuf from %p ...", t->ip());
+		LOG(debug) <<"  stepi out of syscallbuf from "<< t->ip();
 		t->cont_singlestep();
 		assert(t->stopped());
 
-		t->get_siginfo(&tmp_si);
+		if (!t->get_siginfo(&tmp_si)) {
+			goto happy_place;
+		}
 		is_syscall = seems_to_be_syscallbuf_syscall_trap(&tmp_si);
 
 		if (!is_syscall && !is_trace_trap(&tmp_si)) {
 			if (HPC_TIME_SLICE_SIGNAL == tmp_si.si_signo) {
-				debug("  ignoring SIG_TIMESLICE");
+				LOG(debug) <<"  discarding HPC_TIME_SLICE_SIGNAL";
 				continue;
 			}
-			if (HPC_TIME_SLICE_SIGNAL == si->si_signo) {
+			if (SYSCALLBUF_DESCHED_SIGNAL == tmp_si.si_signo) {
+				if (!t->is_untraced_syscall()) {
+					// The performance counter is disarmed but
+					// maybe this signal could already be pending.
+					// We can ignore this signal.
+					LOG(debug) <<"  discarding SYSCALLBUF_DESCHED_SIGNAL";
+					continue;
+				}
+				// TODO what if we get a signal just before
+				// entering an untraced syscall, we step into
+				// that syscall, and then it blocks? We'll
+				// reach this point, but what should we do?
+				// We can probably treat this as a happy_place
+				// and just deliver our original signal, but
+				// we'll need logic similar to handle_desched_event.
+			}
+			// In a desperate effort to avoid dying, discard
+			// an ignored signal. This will preserve tracee
+			// behavior. It means such signals won't be observed
+			// in a debugger, but that will hardly ever be
+			// important.
+			if (HPC_TIME_SLICE_SIGNAL == si->si_signo ||
+				t->is_sig_ignored(si->si_signo)) {
 				memcpy(si, &tmp_si, sizeof(*si));
-				debug("  upgraded delivery of SIG_TIMESLICE to %s",
-				      signalname(si->si_signo));
+				LOG(debug) <<"  upgraded delivery of HPC_TIME_SLICE_SIGNAL to "
+					   << signalname(si->si_signo);
 				handle_siginfo(t, si);
 				return -1;
 			}
+			// In another desperate effort to avoid dying, discard
+			// the new signal if it's ignored.
+			if (t->is_sig_ignored(tmp_si.si_signo)) {
+				LOG(debug) <<"  discarding ignored signal "<< tmp_si.si_signo;
+				continue;
+			}
 
-			assert_exec(t, 0,
-				    "TODO: support multiple pending signals; received %s (code: %d) at $ip:%p while trying to deliver %s (code: %d)",
-				    signalname(tmp_si.si_signo),
-				    tmp_si.si_code, t->ip(),
-				    signalname(si->si_signo), si->si_code);
+			ASSERT(t, false)
+				<< "TODO: support multiple pending signals; received "
+				<< signalname(tmp_si.si_signo) <<" (code: "
+				<< tmp_si.si_code <<") at $ip:"
+				<< t->ip() <<" while trying to deliver "
+				<< signalname(si->si_signo) <<" (code: "
+				<< si->si_code <<")";
 		}
 		if (!is_syscall) {
 			continue;
 		}
 
-		/* TODO more signals can be delivered while we're
-		 * stepping here too.  Sigh.  See comment above about
-		 * masking signals off.  When we mask off signals, we
-		 * won't need to disarm the desched event, but we will
-		 * need to handle spurious desched notifications. */
-		if (t->is_desched_event_syscall()) {
-			debug("  stepping over desched-event syscall");
-			/* Finish the syscall. */
-			t->cont_singlestep();
-			if (t->is_arm_desched_event_syscall()) {
-				/* Disarm the event: we don't need or
-				 * want to hear about descheds while
-				 * we're stepping the tracee through
-				 * the syscall wrapper. */
-				disarm_desched_event(t);
-			}
-			/* We don't care about disarm-desched-event
-			 * syscalls; they're irrelevant. */
-		} else {
-			debug("  running wrapped syscall");
-			/* We may have been notified of the signal
-			 * just after arming the event, but just
-			 * before entering the syscall.  So disarm for
-			 * safety. */
-			/* XXX we really should warn about this, but
-			 * it's just too noisy during unit tests.
-			 * Should find a better way to choose mode. */
-			/*log_warn("Disabling context-switching for possibly-blocking syscall (%s); deadlock may follow",
-			  syscallname(regs->orig_eax));*/
-			disarm_desched_event(t);
-			/* And (hopefully!) finish the syscall. */
-			t->cont_singlestep();
-		}
+		LOG(debug) <<"  running wrapped syscall";
+		/* Finish the syscall. */
+		t->cont_singlestep();
 	}
 
 happy_place:
@@ -587,9 +593,8 @@ happy_place:
 
 static void handle_siginfo(Task* t, siginfo_t* si)
 {
-	debug("%d: handling signal %s (pevent: %d, event: %s)",
-	      t->tid, signalname(si->si_signo),
-	      t->ptrace_event(), event_name(t->ev()));
+	LOG(debug) << t->tid <<": handling signal "<< signalname(si->si_signo)
+		   <<" (pevent: "<< t->ptrace_event() <<", event: "<< t->ev();
 
 	/* We have to check for a desched event first, because for
 	 * those we *do not* want to (and cannot, most of the time)
@@ -632,7 +637,10 @@ void handle_signal(Task* t, siginfo_t* si)
 
 	siginfo_t local_si;
 	if (!si) {
-		t->get_siginfo(&local_si);
+		if (!t->get_siginfo(&local_si)) {
+			// task died. Nothing to do here.
+			return;
+		}
 		si = &local_si;
 	}
 	return handle_siginfo(t, si);
