@@ -1,7 +1,7 @@
 /* -*- Mode: C++; tab-width: 8; c-basic-offset: 8; indent-tabs-mode: t; -*- */
 
-#ifndef PROCESS_SYSCALL_H_
-#define PROCESS_SYSCALL_H_
+#ifndef RR_PROCESS_SYSCALL_H_
+#define RR_PROCESS_SYSCALL_H_
 
 #include "types.h"
 #include "util.h"
@@ -24,7 +24,7 @@ void rec_before_record_syscall_entry(Task* t, int syscallno);
  * overly general mechanism that's used for FUTEX_LOCK_PI.  If you're
  * not FUTEX_LOCK_PI, you probably shouldn't be using this.
  */
-int rec_prepare_syscall(Task* t, void** kernel_sync_addr, uint32_t* sync_val);
+int rec_prepare_syscall(Task* t);
 /**
  * Prepare |t| for its current syscall event to be interrupted and
  * possibly restarted.
@@ -36,4 +36,4 @@ void rec_prepare_restart_syscall(Task* t);
  */
 void rec_process_syscall(Task* t);
 
-#endif /* PROCESS_SYSCALL_H_ */
+#endif /* RR_PROCESS_SYSCALL_H_ */
