@@ -23,22 +23,22 @@
 
 #include "preload/syscall_buffer.h"
 
-#include "dbg.h"
 #include "hpc.h"
+#include "log.h"
 #include "record_signal.h"
 #include "record_syscall.h"
 #include "recorder_sched.h"
+#include "session.h"
 #include "task.h"
 #include "trace.h"
 #include "util.h"
 
 using namespace std;
 
-static string exe_image;
-// NB: we currently intentionally leak the constituent strings in
-// these arrays.
-static CharpVector arg_v;
-static CharpVector env_p;
+// Global state of this recording session.
+static RecordSession::shr_ptr session;
+// Args/env for the initial tracee.
+struct args_env ae;
 
 /* Nonzero when it's safe to deliver signals, namely, when the initial
  * tracee has exec()'d the tracee image.  Before then, the address
@@ -47,47 +47,6 @@ static CharpVector env_p;
  * the signal. */
 static int can_deliver_signals;
 
-static void copy_argv(int argc, char* argv[])
-{
-	for (int i = 0; i < argc; ++i) {
-		arg_v.push_back(strdup(argv[i]));
-	}
-	arg_v.push_back(NULL);
-}
-
-static void copy_envp(char** envp)
-{
-	int i = 0, preload_index = -1;
-	for (i = 0; envp[i]; ++i) {
-		env_p.push_back(strdup(envp[i]));
-		if (envp[i] == strstr(envp[i], "LD_PRELOAD=")) {
-			preload_index = i;
-		}
-	}
-	// LD_PRELOAD the syscall interception lib
-	if (rr_flags()->syscall_buffer_lib_path) {
-		string ld_preload = "LD_PRELOAD=";
-		// Our preload lib *must* come first
-		ld_preload += rr_flags()->syscall_buffer_lib_path;
-		if (preload_index >= 0) {
-			const char* old_preload =
-				strchr(envp[preload_index], '=') + 1;
-			assert(old_preload);
-			// Honor old preloads too.  this may cause
-			// problems, but only in those libs, and
-			// that's the user's problem.
-			ld_preload += ":";
-			ld_preload += old_preload;
-		} else {
-			/* Or if this is a new key/value, "allocate"
-			 * an index for it */
-			preload_index = i++;
-		}
-		env_p.push_back(strdup(ld_preload.c_str()));
-	}
-	env_p.push_back(NULL);
-}
-
 /**
  * Create a pulseaudio client config file with shm disabled.  That may
  * be the cause of a mysterious divergence.  Return an envpair to set
@@ -98,11 +57,15 @@ static string create_pulseaudio_config()
 	// TODO let PULSE_CLIENTCONFIG env var take precedence.
 	static const char pulseaudio_config_path[] = "/etc/pulse/client.conf";
 	if (access(pulseaudio_config_path, R_OK)) {
-		fatal("Can't file pulseaudio config at %s.", pulseaudio_config_path);
+		// Assume pulseaudio isn't installed
+		return "";
 	}
-	char tmp[] = "rr-pulseaudio-client-conf-XXXXXX";
+	char tmp[] = "/tmp/rr-pulseaudio-client-conf-XXXXXX";
 	int fd = mkstemp(tmp);
+	fcntl(fd, F_SETFD, FD_CLOEXEC);
 	unlink(tmp);
+	// The fd is deliberately leaked so that the /proc/fd link below works
+	// indefinitely. But we stop it leaking into tracee processes.
 
 	stringstream procfile;
 	procfile << "/proc/" << getpid() << "/fd/" << fd;
@@ -111,16 +74,16 @@ static string create_pulseaudio_config()
 	    
 	int status = system(cmd.str().c_str());
 	if (-1 == status || !WIFEXITED(status) || 0 != WEXITSTATUS(status)) {
-		fatal("The command '%s' failed.", cmd.str().c_str());
+		FATAL() <<"The command '"<< cmd.str() <<"' failed.";
 	}
 	if (-1 == lseek(fd, 0, SEEK_END)) {
-		fatal("Failed to seek to end of file.");
+		FATAL() <<"Failed to seek to end of file.";
 	}
 	char disable_shm[] = "disable-shm = true\n";
 	ssize_t nwritten = write(fd, disable_shm, sizeof(disable_shm) - 1);
 	if (nwritten != sizeof(disable_shm) - 1) {
-		fatal("Failed to append '%s' to %s",
-		      disable_shm, procfile.str().c_str());
+		FATAL() <<"Failed to append '"<< disable_shm <<"' to "
+			<< procfile.str();
 	}
 	stringstream envpair;
 	envpair << "PULSE_CLIENTCONFIG=" << procfile.str();
@@ -147,22 +110,20 @@ static void ensure_preload_lib_will_load(const char* rr_exe,
 	pid_t child = fork();
 	if (0 == child) {
 		execvpe(rr_exe, argv, ep.data());
-		fatal("Failed to exec %s", rr_exe);
+		FATAL() <<"Failed to exec "<< rr_exe;
 	}
 	int status;
 	pid_t ret = waitpid(child, &status, 0);
 	if (ret != child) {
-		fatal("Failed to wait for %s child", rr_exe);
+		FATAL() <<"Failed to wait for "<< rr_exe <<" child";
 	}
 	if (!WIFEXITED(status) || 0 != WEXITSTATUS(status)) {
+		const struct flags* flags = rr_flags();
 		fprintf(stderr,
 "\n"
 "rr: error: Unable to preload the '%s' library.\n"
-"  Ensure that the library is in your LD_LIBRARY_PATH.  If you installed rr\n"
-"  from a distribution package, then the package or your system was not\n"
-"  configured correctly.\n"
 "\n",
-			SYSCALLBUF_LIB_FILENAME);
+			flags->syscall_buffer_lib_path.c_str());
 		exit(EX_CONFIG);
 	}
 }
@@ -174,29 +135,30 @@ static void handle_ptrace_event(Task** tp)
 	/* handle events */
 	int event = t->ptrace_event();
 	if (event != PTRACE_EVENT_NONE) {
-		debug("  %d: handle_ptrace_event %d: event %s",
-		      t->tid, event, event_name(t->ev()));
+		LOG(debug) <<"  "<< t->tid <<": handle_ptrace_event "
+			   << event <<": event "<< t->ev();
 	}
 	switch (event) {
 
 	case PTRACE_EVENT_NONE:
 	case PTRACE_EVENT_STOP:
+	case PTRACE_EVENT_SECCOMP_OBSOLETE:
+	case PTRACE_EVENT_SECCOMP:
 		break;
 
 	case PTRACE_EVENT_CLONE:
 	case PTRACE_EVENT_FORK: {
-		int new_tid = t->get_ptrace_eventmsg();
-		void* stack = (void*)t->regs().ecx;
-		void* ctid = (void*)t->regs().edi;
+		int new_tid = t->get_ptrace_eventmsg_pid();
+		void* stack = (void*)t->regs().arg2();
+		void* tls = (void*)t->regs().arg4();
+		void* ctid = (void*)t->regs().arg5();
 		// fork and can never share these resources, only
 		// copy, so the flags here aren't meaningful for it.
-		int flags_arg = (SYS_clone == t->regs().orig_eax) ?
-				t->regs().ebx : 0;
-		Task* new_task = t->clone(clone_flags_to_task_flags(flags_arg),
-					  stack, ctid, new_tid);
-		// Wait until the new task is ready.
-		new_task->wait();
-		start_hpc(new_task, rr_flags()->max_rbc);
+		unsigned long flags_arg = (SYS_clone == t->regs().original_syscallno()) ?
+			t->regs().arg1() : 0;
+		t->session().clone(
+			t, clone_flags_to_task_flags(flags_arg),
+			stack, tls, ctid, new_tid);
 		// Skip past the ptrace event.
 		t->cont_syscall();
 		assert(t->pending_sig() == 0);
@@ -211,7 +173,7 @@ static void handle_ptrace_event(Task** tp)
 
 		t->push_event(SyscallEvent(SYS_execve));
 		t->ev().Syscall().state = ENTERING_SYSCALL;
-		record_event(t);
+		t->record_current_event();
 		t->pop_syscall();
 
 		// Skip past the ptrace event.
@@ -221,17 +183,15 @@ static void handle_ptrace_event(Task** tp)
 	}
 
 	case PTRACE_EVENT_EXIT: {
-		if (EV_SYSCALL == t->ev().type()
-		    && SYS_exit_group == t->ev().Syscall().no
-		    && t->task_group()->task_set().size() > 1) {
-			log_warn("exit_group() with > 1 task; may misrecord CLONE_CHILD_CLEARTID memory race");
+		if (t->stable_exit) {
+			LOG(debug) <<"stable exit";
+		} else {
+			LOG(warn) <<"unstable exit; may misrecord CLONE_CHILD_CLEARTID memory race";
 			t->destabilize_task_group();
 		}
 
 		EventType ev = t->unstable ? EV_UNSTABLE_EXIT : EV_EXIT;
-		t->push_event(Event(ev, HAS_EXEC_INFO));
-		record_event(t);
-		t->pop_event(ev);
+		t->record_event(Event(ev, NO_EXEC_INFO));
 
 		rec_sched_deregister_thread(tp);
 		t = *tp;
@@ -240,15 +200,17 @@ static void handle_ptrace_event(Task** tp)
 	case PTRACE_EVENT_VFORK:
 	case PTRACE_EVENT_VFORK_DONE:
 	default:
-		fatal("Unhandled ptrace event %s(%d)",
-		      ptrace_event_name(event), event);
+		FATAL() <<"Unhandled ptrace event "<< ptrace_event_name(event)
+			<<"("<< event <<")";
 		break;
 	}
 }
 
-#define debug_exec_state(_msg, _t)					\
-	debug(_msg ": status=0x%x pevent=%d",				\
-	      (_t)->status(), (_t)->ptrace_event())
+static void debug_exec_state(const char* msg, Task* t)
+{
+	LOG(debug) << msg <<": status="<< HEX(t->status())
+		   <<" pevent="<< t->ptrace_event();
+}
 
 enum { DEFAULT_CONT = 0, FORCE_SYSCALL = 1 };
 static void task_continue(Task* t, int force_cont, int sig)
@@ -256,11 +218,12 @@ static void task_continue(Task* t, int force_cont, int sig)
 	bool may_restart = t->at_may_restart_syscall();
 
 	if (sig) {
-		debug("  delivering %s to %d", signalname(sig), t->tid);
+		LOG(debug) <<"  delivering "<< signalname(sig)
+			   <<" to "<< t->tid;
 	}
 	if (may_restart && t->seccomp_bpf_enabled) {
-		debug("  PTRACE_SYSCALL to possibly-restarted %s",
-		      syscallname(t->ev().Syscall().no));
+		LOG(debug) <<"  PTRACE_SYSCALL to possibly-restarted "
+			   << t->ev();
 	}
 
 	if (!t->seccomp_bpf_enabled
@@ -270,7 +233,7 @@ static void task_continue(Task* t, int force_cont, int sig)
 		 * syscall_buffer lib in the child, therefore we must
 		 * record in the traditional way (with PTRACE_SYSCALL)
 		 * until it is installed. */
-		t->cont_syscall_nonblocking(sig);
+		t->cont_syscall_nonblocking(sig, rr_flags()->max_rbc);
 	} else {
 		/* When the seccomp filter is on, instead of capturing
 		 * syscalls by using PTRACE_SYSCALL, the filter will
@@ -282,7 +245,7 @@ static void task_continue(Task* t, int force_cont, int sig)
 		 * process to continue to the actual entry point of
 		 * the syscall (using cont_syscall_block()) and then
 		 * using the same logic as before. */
-		t->cont_nonblocking(sig);
+		t->cont_nonblocking(sig, rr_flags()->max_rbc);
 	}
 }
 
@@ -301,7 +264,7 @@ static bool resume_execution(Task* t, int need_task_continue,
 	if (need_task_continue) {
 		task_continue(t, force_cont, /*no sig*/0);
 		if (!t->wait()) {
-			debug("  waitpid() interrupted");
+			LOG(debug) <<"  waitpid() interrupted";
 			return false;
 		}
 	}
@@ -309,7 +272,7 @@ static bool resume_execution(Task* t, int need_task_continue,
 	if (t->is_ptrace_seccomp_event()) {
 		t->seccomp_bpf_enabled = true;
 		/* See long comments above. */
-		debug("  (skipping past seccomp-bpf trap)");
+		LOG(debug) <<"  (skipping past seccomp-bpf trap)";
 		return resume_execution(t, NEED_TASK_CONTINUE, FORCE_SYSCALL);
 	}
 	return true;
@@ -325,7 +288,7 @@ static void disarm_desched(Task* t)
 {
 	int old_sig = 0;
 
-	debug("desched: DISARMING_DESCHED_EVENT");
+	LOG(debug) <<"desched: DISARMING_DESCHED_EVENT";
 	/* TODO: send this through main loop. */
 	/* TODO: mask off signals and avoid this loop. */
 	do {
@@ -348,11 +311,11 @@ static void disarm_desched(Task* t)
 			continue;
 		}
 		if (sig && sig == old_sig) {
-			debug("  coalescing pending %s", signalname(sig));
+			LOG(debug) <<"  coalescing pending "<< signalname(sig);
 			continue;
 		}
 		if (sig) {
-			debug("  %s now pending", signalname(sig));
+			LOG(debug) <<"  "<< signalname(sig) <<" now pending";
 			t->stash_sig();
 		}
 	} while (!t->is_disarm_desched_event_syscall());
@@ -367,15 +330,13 @@ static void desched_state_changed(Task* t)
 {
 	switch (t->ev().Desched().state) {
 	case IN_SYSCALL:
-		debug("desched: IN_SYSCALL");
+		LOG(debug) <<"desched: IN_SYSCALL";
 		/* We need to ensure that the syscallbuf code doesn't
 		 * try to commit the current record; we've already
 		 * recorded that syscall.  The following event sets
 		 * the abort-commit bit. */
-		t->push_event(Event(EV_SYSCALLBUF_ABORT_COMMIT, NO_EXEC_INFO));
 		t->syscallbuf_hdr->abort_commit = 1;
-		record_event(t);
-		t->pop_event(EV_SYSCALLBUF_ABORT_COMMIT);
+		t->record_event(Event(EV_SYSCALLBUF_ABORT_COMMIT, NO_EXEC_INFO));
 
 		t->ev().Desched().state = DISARMING_DESCHED_EVENT;
 		/* fall through */
@@ -383,19 +344,17 @@ static void desched_state_changed(Task* t)
 		disarm_desched(t);
 
 		t->ev().Desched().state = DISARMED_DESCHED_EVENT;
-		record_event(t);
+		t->record_current_event();
 		t->pop_desched();
 
 		/* The tracee has just finished sanity-checking the
 		 * aborted record, and won't touch the syscallbuf
 		 * during this (aborted) transaction again.  So now is
 		 * a good time for us to reset the record counter. */
-		t->push_event(Event(EV_SYSCALLBUF_RESET, NO_EXEC_INFO));
 		t->syscallbuf_hdr->num_rec_bytes = 0;
 		t->delay_syscallbuf_reset = 0;
 		t->delay_syscallbuf_flush = 0;
-		record_event(t);
-		t->pop_event(EV_SYSCALLBUF_RESET);
+		t->record_event(Event(EV_SYSCALLBUF_RESET, NO_EXEC_INFO));
 		// We were just descheduled for potentially a long
 		// time, and may have just had a signal become
 		// pending.  Ensure we get another chance to run.
@@ -403,23 +362,21 @@ static void desched_state_changed(Task* t)
 		return;
 	}
 	default:
-		fatal("Unhandled desched state");
+		FATAL() <<"Unhandled desched state";
 	}
 }
 
 static void syscall_not_restarted(Task* t)
 {
-	debug("  %d: popping abandoned interrupted %s; pending events:",
-	      t->tid, syscallname(t->ev().Syscall().no));
+	LOG(debug) <<"  "<< t->tid <<": popping abandoned interrupted "
+		   << t->ev() <<"; pending events:";
 #ifdef DEBUGTAG
-	log_pending_events(t);
+	t->log_pending_events();
 #endif
 	t->pop_syscall_interruption();
 
-	t->push_event(Event(EV_INTERRUPTED_SYSCALL_NOT_RESTARTED,
-			    NO_EXEC_INFO));
-	record_event(t);
-	t->pop_event(EV_INTERRUPTED_SYSCALL_NOT_RESTARTED);
+	t->record_event(Event(EV_INTERRUPTED_SYSCALL_NOT_RESTARTED,
+			      NO_EXEC_INFO));
 }
 
 /**
@@ -432,12 +389,15 @@ static void syscall_not_restarted(Task* t)
  */
 static int maybe_restart_syscall(Task* t)
 {
-	if (SYS_restart_syscall == t->regs().orig_eax) {
-		debug("  %d: SYS_restart_syscall'ing %s",
-		      t->tid, syscallname(t->ev().Syscall().no));
+	if (SYS_restart_syscall == t->regs().original_syscallno()) {
+		LOG(debug) <<"  "<< t->tid <<": SYS_restart_syscall'ing "
+			   << t->ev();
 	}
 	if (t->is_syscall_restart()) {
 		t->ev().transform(EV_SYSCALL);
+		Registers regs = t->regs();
+		regs.set_original_syscallno(t->ev().Syscall().regs.original_syscallno());
+		t->set_regs(regs);
 		return 1;
 	}
 	if (EV_SYSCALL_INTERRUPTION == t->ev().type()) {
@@ -459,7 +419,7 @@ static void maybe_discard_syscall_interruption(Task* t, int ret)
 		/* We currently don't track syscalls interrupted with
 		 * ERESTARTSYS or ERESTARTNOHAND, so it's possible for
 		 * a sigreturn not to affect the event stack. */
-		debug("  (no interrupted syscall to retire)");
+		LOG(debug) <<"  (no interrupted syscall to retire)";
 		return;
 	}
 
@@ -467,9 +427,10 @@ static void maybe_discard_syscall_interruption(Task* t, int ret)
 	if (0 > ret) {
 		syscall_not_restarted(t);
 	} else if (0 < ret) {
-		assert_exec(t, syscallno == ret,
-			    "Interrupted call was %s, and sigreturn claims to be restarting %s",
-			    syscallname(syscallno), syscallname(ret));
+		ASSERT(t, syscallno == ret)
+			<<"Interrupted call was "<< t->syscallname(syscallno)
+			<<" and sigreturn claims to be restarting "
+			<< t->syscallname(ret);
 	}
 }
 
@@ -491,17 +452,12 @@ static void syscall_state_changed(Task* t, int by_waitpid)
 			t->ev().Syscall().regs = t->regs();
 		}
 
-		void* sync_addr = nullptr;
-		uint32_t sync_val;
-		t->switchable = rec_prepare_syscall(t, &sync_addr, &sync_val);
+		t->switchable = rec_prepare_syscall(t);
 
 		// Resume the syscall execution in the kernel context.
 		t->cont_syscall_nonblocking();
 		debug_exec_state("after cont", t);
 
-		if (sync_addr) {
-			t->futex_wait(sync_addr, sync_val);
-		}
 		t->ev().Syscall().state = PROCESSING_SYSCALL;
 		return;
 	}
@@ -511,9 +467,9 @@ static void syscall_state_changed(Task* t, int by_waitpid)
 		assert(by_waitpid);
 		// Linux kicks tasks out of syscalls before delivering
 		// signals.
-		assert_exec(t, !t->pending_sig(),
-			    "Signal %s pending while %d in syscall???",
-			    signalname(t->pending_sig()), t->tid);
+		ASSERT(t, !t->pending_sig())
+			<< "Signal "<< signalname(t->pending_sig())
+			<<" pending while in syscall???";
 
 		t->ev().Syscall().state = EXITING_SYSCALL;
 		t->switchable = 0;
@@ -528,7 +484,7 @@ static void syscall_state_changed(Task* t, int by_waitpid)
 
 		assert(t->pending_sig() == 0);
 
-		retval = t->regs().eax;
+		retval = t->regs().syscall_result();
 
 		// sigreturn is a special snowflake, because it
 		// doesn't actually return.  Instead, it undoes the
@@ -538,15 +494,13 @@ static void syscall_state_changed(Task* t, int by_waitpid)
 		// interrupted-syscall event.
 		if (SYS_sigreturn == syscallno
 		    || SYS_rt_sigreturn == syscallno) {
-			assert(t->regs().orig_eax == -1);
-			record_event(t);
+			assert(t->regs().original_syscallno() == -1);
+			t->record_current_event();
 			t->pop_syscall();
 
 			// We've finished processing this signal now.
 			t->pop_signal_handler();
-			t->push_event(Event(EV_EXIT_SIGHANDLER, NO_EXEC_INFO));
-			record_event(t);
-			t->pop_event(EV_EXIT_SIGHANDLER);
+			t->record_event(Event(EV_EXIT_SIGHANDLER, NO_EXEC_INFO));
 
 			maybe_discard_syscall_interruption(t, retval);
 			// XXX probably not necessary to make the
@@ -555,24 +509,25 @@ static void syscall_state_changed(Task* t, int by_waitpid)
 			return;
 		}
 
-		assert_exec(t, (-ENOSYS != retval
-				|| (0 > syscallno
-				    || SYS_rrcall_init_buffers == syscallno
-				    || SYS_rrcall_monkeypatch_vdso == syscallno
-				    || SYS_clone == syscallno
-				    || SYS_exit_group == syscallno
-				    || SYS_exit == syscallno)),
-			    "Exiting syscall %s, but retval is -ENOSYS, usually only seen at entry",
-			    syscallname(syscallno));
+		ASSERT(t, (-ENOSYS != retval
+			   || (0 > syscallno
+			       || SYS_rrcall_init_buffers == syscallno
+			       || SYS_rrcall_monkeypatch_vdso == syscallno
+			       || SYS_clone == syscallno
+			       || SYS_exit_group == syscallno
+			       || SYS_exit == syscallno
+			       || SYS__sysctl == syscallno)))
+			<< "Exiting syscall "<< t->syscallname(syscallno)
+			<<" but retval is -ENOSYS, usually only seen at entry";
 
-		debug("  orig_eax:%ld (%s); eax:%ld",
-		      t->regs().orig_eax, syscallname(syscallno),
-		      t->regs().eax);
+		LOG(debug) <<"  original_syscallno:"<< t->regs().original_syscallno()
+			   <<" ("<< t->syscallname(syscallno) <<"); return val:"
+			   << t->regs().syscall_result();
 
 		/* a syscall_restart ending is equivalent to the
 		 * restarted syscall ending */
 		if (t->ev().Syscall().is_restart) {
-			debug("  exiting restarted %s", syscallname(syscallno));
+			LOG(debug) <<"  exiting restarted "<< t->syscallname(syscallno);
 		}
 
 		/* TODO: is there any reason a restart_syscall can't
@@ -587,12 +542,13 @@ static void syscall_state_changed(Task* t, int by_waitpid)
 		 * restart_syscall */
 		if (!may_restart) {
 			rec_process_syscall(t);
-			if (rr_flags()->check_cached_mmaps) {
+			if (t->session().can_validate()
+			    && rr_flags()->check_cached_mmaps) {
 				t->vm()->verify(t);
 			}
 		} else {
-			debug("  may restart %s (from retval %d)",
-			      syscallname(syscallno), retval);
+			LOG(debug) <<"  may restart "<< t->syscallname(syscallno)
+				   <<" (from retval "<< retval <<")";
 
 			rec_prepare_restart_syscall(t);
 			/* If we may restart this syscall, we've most
@@ -601,11 +557,11 @@ static void syscall_state_changed(Task* t, int by_waitpid)
 			 * want to record those fudged registers,
 			 * because scratch doesn't exist in replay.
 			 * So cover our tracks here. */
-			struct user_regs_struct r = t->regs();
+			Registers r = t->regs();
 			copy_syscall_arg_regs(&r, &t->ev().Syscall().regs);
 			t->set_regs(r);
 		}
-		record_event(t);
+		t->record_current_event();
 
 		/* If we're not going to restart this syscall, we're
 		 * done with it.  But if we are, "freeze" it on the
@@ -614,7 +570,7 @@ static void syscall_state_changed(Task* t, int by_waitpid)
 		if (!may_restart) {
 			t->pop_syscall();
 			if (EV_DESCHED == t->ev().type()) {
-				debug("  exiting desched critical section");
+				LOG(debug) <<"  exiting desched critical section";
 				desched_state_changed(t);
 			}
 		} else {
@@ -627,7 +583,7 @@ static void syscall_state_changed(Task* t, int by_waitpid)
 	}
 
 	default:
-		fatal("Unknown exec state %d", t->ev().Syscall().state);
+		FATAL() <<"Unknown exec state "<< t->ev().Syscall().state;
 	}
 }
 
@@ -639,9 +595,7 @@ static void syscall_state_changed(Task* t, int by_waitpid)
 static void maybe_reset_syscallbuf(Task* t)
 {
 	if (t->flushed_syscallbuf && !t->delay_syscallbuf_reset) {
-		t->push_event(Event(EV_SYSCALLBUF_RESET, NO_EXEC_INFO));
-		record_event(t);
-		t->pop_event(EV_SYSCALLBUF_RESET);
+		t->record_event(Event(EV_SYSCALLBUF_RESET, NO_EXEC_INFO));
 	}
 	/* Any code that sets |delay_syscallbuf_reset| is responsible
 	 * for recording its own SYSCALLBUF_RESET event at a
@@ -655,29 +609,32 @@ static void check_rbc(Task* t)
 	if (can_deliver_signals || SYS_write != t->ev().Syscall().no) {
 		return;
 	}
-	int fd = t->regs().ebx;
+	int fd = t->regs().arg1_signed();
+	if (-1 != fd && rr_flags()->force_things) {
+		LOG(warn) <<"Unexpected write("<< fd <<") call";
+		return;
+	}
 	if (-1 != fd) {
 		fprintf(stderr,
 "\n"
 "rr: error:\n"
 "  Unexpected `write(%d, ...)' call from first tracee process.\n"
-"  Most likely, the executable image `%s' doesn't exist or isn't\n"
-"  in your $PATH.  Terminating recording.\n"
+"  Most likely, the executable image `%s' is 64-bit, doesn't exist, or\n"
+"  isn't in your $PATH.  Terminating recording.\n"
 "\n",
-			fd, exe_image.c_str());
+			fd, ae.exe_image.c_str());
 		terminate_recording(t);
-		return;
 	}
 
-	int64_t rbc = read_rbc(t->hpc);
-	debug("rbc on entry to dummy write: %lld", rbc);
+	int64_t rbc = t->rbc_count();
+	LOG(debug) <<"rbc on entry to dummy write: " << rbc;
 	if (!(rbc > 0)) {
 		fprintf(stderr,
 "\n"
 "rr: internal recorder error:\n"
 "  Retired-branch counter doesn't seem to be working.  Are you perhaps\n"
 "  running rr in a VM but didn't enable perf-counter virtualization?\n");
-		exit(EX_UNAVAILABLE);
+		terminate_recording(t, EX_UNAVAILABLE);
 	}
 }
 
@@ -700,14 +657,12 @@ static bool signal_state_changed(Task* t, int by_waitpid)
 
 		// This event is used by the replayer to advance to
 		// the point of signal delivery.
-		record_event(t);
-		reset_hpc(t, rr_flags()->max_rbc);
-
+		t->record_current_event();
 		t->ev().transform(EV_SIGNAL_DELIVERY);
 		ssize_t sigframe_size;
 		if (t->signal_has_user_handler(sig)) {
-			debug("  %d: %s has user handler", t->tid,
-			      signalname(sig));
+			LOG(debug) <<"  "<< t->tid <<": "<< signalname(sig)
+				   <<" has user handler";
 
 			if (!t->cont_singlestep(sig)) {
 				return false;
@@ -746,19 +701,19 @@ static bool signal_state_changed(Task* t, int by_waitpid)
 			t->signal_delivered(sig);
 			t->ev().Signal().delivered = 1;
 		} else {
-			debug("  %d: no user handler for %s", t->tid,
-			      signalname(sig));
+			LOG(debug) <<"  "<< t->tid <<": no user handler for "
+				   << signalname(sig);
 			sigframe_size = 0;
 		}
 
 		// We record this data regardless to simplify replay.
-		record_child_data(t, sigframe_size, t->sp());
+		t->record_remote(t->sp(), sigframe_size);
 
 		// This event is used by the replayer to set up the
 		// signal handler frame, or to record the resulting
 		// state of the stepi if there wasn't a signal
 		// handler.
-		record_event(t);
+		t->record_current_event();
 
 		// If we didn't set up the sighandler frame, we need
 		// to ensure that this tracee is scheduled next so
@@ -775,7 +730,7 @@ static bool signal_state_changed(Task* t, int by_waitpid)
 		if (!t->ev().Signal().delivered) {
 			task_continue(t, DEFAULT_CONT, sig);
 			if (possibly_destabilizing_signal(t, sig)) {
-				log_warn("Delivered core-dumping signal; may misrecord CLONE_CHILD_CLEARTID memory race");
+				LOG(warn) <<"Delivered core-dumping signal; may misrecord CLONE_CHILD_CLEARTID memory race";
 				t->destabilize_task_group();
 				t->switchable = 1;
 			}
@@ -793,7 +748,7 @@ static bool signal_state_changed(Task* t, int by_waitpid)
 		return true;
 
 	default:
-		fatal("Unhandled signal state %d", t->ev().type());
+		FATAL() <<"Unhandled signal state " << t->ev().type();
 		return false;	// not reached
 	}
 }
@@ -809,12 +764,19 @@ static void runnable_state_changed(Task* t)
 	// to allow switching the context.
 	t->switchable = 0;
 
+	if (t->ptrace_event()) {
+		// A ptrace event arrived. The steps below are irrelevant
+		// and potentially wrong because no ev() was pushed.
+		return;
+	}
+
 	siginfo_t* si = nullptr;
 	siginfo_t stash;
 	if (t->has_stashed_sig()) {
 		stash = t->pop_stash_sig();
 		si = &stash;
-		debug("pulled %s out of stash", signalname(t->pending_sig()));
+		LOG(debug) <<"pulled "<< signalname(t->pending_sig())
+			   <<" out of stash";
 	}
 
 	if (t->pending_sig() && can_deliver_signals) {
@@ -829,8 +791,8 @@ static void runnable_state_changed(Task* t)
 		//
 		// This doesn't really occur in practice, only in
 		// tests that force a degenerately low time slice.
-		log_warn("Dropping %s because it can't be delivered yet",
-			 signalname(t->pending_sig()));
+		LOG(warn) <<"Dropping "<< signalname(t->pending_sig())
+			  <<" because it can't be delivered yet";
 		// No events to be recorded, so no syscallbuf updates
 		// needed.
 		return;
@@ -842,7 +804,7 @@ static void runnable_state_changed(Task* t)
 		break;
 	case EV_SEGV_RDTSC:
 	case EV_SCHED:
-		record_event(t);
+		t->record_current_event();
 		t->pop_event(t->ev().type());
 		t->switchable = 1;
 		break;
@@ -855,20 +817,18 @@ static void runnable_state_changed(Task* t)
 	case EV_SYSCALL_INTERRUPTION:
 		// We just entered a syscall.
 		if (!maybe_restart_syscall(t)) {
-			t->push_event(SyscallEvent(t->regs().orig_eax));
+			t->push_event(SyscallEvent(t->regs().original_syscallno()));
 			rec_before_record_syscall_entry(t, t->ev().Syscall().no);
 		}
-		assert_exec(t, EV_SYSCALL == t->ev().type(),
-			    "Should be at syscall event.");
+		ASSERT(t, EV_SYSCALL == t->ev().type());
 		check_rbc(t);
 		t->ev().Syscall().state = ENTERING_SYSCALL;
-		record_event(t);
+		t->record_current_event();
 		break;
 
 	default:
-		assert_exec(t, false,
-			    "%s can't be on event stack at start of new event",
-			    t->ev().str().c_str());
+		ASSERT(t, false)
+			<< t->ev() <<" can't be on event stack at start of new event";
 		break;
 	}
 	maybe_reset_syscallbuf(t);
@@ -886,10 +846,10 @@ static bool term_request;
 static void handle_termsig(int sig)
 {
 	if (term_request) {
-		fatal("Received termsig while an earlier one was pending.  We're probably wedged.");
+		FATAL() <<"Received termsig while an earlier one was pending.  We're probably wedged.";
 	}
-	log_info("Received termsig %s, requesting shutdown ...\n",
-		 signalname(sig));
+	LOG(info) <<"Received termsig "<< signalname(sig)
+		  <<", requesting shutdown ...\n";
 	term_request = true;
 }
 
@@ -908,62 +868,133 @@ static void install_termsig_handlers(void)
 static void maybe_process_term_request(Task* t)
 {
 	if (term_request) {
+		t->maybe_flush_syscallbuf();
 		terminate_recording(t);
 	}
 }
 
-void record(const char* rr_exe, int argc, char* argv[], char** envp)
+/**
+ * Pick a CPU at random to bind to, unless --cpu-unbound has been given,
+ * in which case we return -1.
+ */
+static int choose_cpu()
 {
-	log_info("Start recording...");
+	if (rr_flags()->cpu_unbound) {
+		return -1;
+	}
 
-	exe_image = argv[0];
-	copy_argv(argc, argv);
-	copy_envp(envp);
-	rec_setup_trace_dir();
+	// Pin tracee tasks to logical CPU 0, both in
+	// recording and replay.  Tracees can see which HW
+	// thread they're running on by asking CPUID, and we
+	// don't have a way to emulate it yet.  So if a tracee
+	// happens to be scheduled on a different core in
+	// recording than replay, it can diverge.  (And
+	// indeed, has been observed to diverge in practice,
+	// in glibc.)
+	//
+	// Note that we will pin both the tracee processes *and*
+	// the tracer process.  This ends up being a tidy
+	// performance win in certain circumstances,
+	// presumably due to cheaper context switching and/or
+	// better interaction with CPU frequency scaling.
+	return random()%get_num_cpus();
+}
+
+int record(const char* rr_exe, int argc, char* argv[], char** envp)
+{
+	LOG(info) <<"Start recording...";
+
+	int bind_to_cpu = choose_cpu();
+
+	char cwd[PATH_MAX] = "";
+	getcwd(cwd, sizeof(cwd));
+
+	ae = args_env(argc, argv, envp, cwd, bind_to_cpu);
+	// LD_PRELOAD the syscall interception lib
+	if (!rr_flags()->syscall_buffer_lib_path.empty()) {
+		// Remove the trailing nullptr.  We'll put it back
+		// after injecting the preload lib into the envp.
+		ae.envp.pop_back();
+		string ld_preload = "LD_PRELOAD=";
+		// Our preload lib *must* come first
+		ld_preload += rr_flags()->syscall_buffer_lib_path;
+		auto it = ae.envp.begin();
+		for (; it != ae.envp.end(); ++it) {
+			if (*it != strstr(*it, "LD_PRELOAD=")) {
+				continue;
+			}
+			const char* old_preload = strchr(*it, '=') + 1;
+			assert(old_preload);
+			// Honor old preloads too.  This may cause
+			// problems, but only in those libs, and
+			// that's the user's problem.
+			ld_preload += ":";
+			ld_preload += old_preload;
+			break;
+		}
+		// The envp vector takes over ownership of this
+		// string.
+		char* new_preload = strdup(ld_preload.c_str());
+		if (it == ae.envp.end()) {
+			ae.envp.push_back(new_preload);
+		} else {
+			free(*it);
+			*it = new_preload;
+		}
+		ae.envp.push_back(nullptr);
+	}
 
 	string env_pair = create_pulseaudio_config();
-	// Intentionally leaked.
-	env_p[env_p.size() - 1] = strdup(env_pair.c_str());
-	env_p.push_back(nullptr);
+	if (!env_pair.empty()) {
+		// The envp vector takes over ownership of this
+		// string.
+		ae.envp.back() = strdup(env_pair.c_str());
+		ae.envp.push_back(nullptr);
+	}
 
-	ensure_preload_lib_will_load(rr_exe, env_p);
+	ensure_preload_lib_will_load(rr_exe, ae.envp);
+	session = RecordSession::create(argv[0]);
 
-	open_trace_files();
-	rec_init_trace_files();
-	record_argv_envp(argc, arg_v.data(), env_p.data());
-	init_libpfm();
+	session->ofstream() << ae;
 
 	install_termsig_handlers();
 
-	Task* t = Task::create(exe_image, arg_v, env_p);
-	start_hpc(t, rr_flags()->max_rbc);
+	Task* t = session->create_task(ae, session);
+	TaskGroup::shr_ptr initial_task_group = t->task_group();
 
-	while (Task::count() > 0) {
+	while (session->tasks().size() > 0) {
 		int by_waitpid;
 
 		maybe_process_term_request(t);
 
-		Task* next = rec_sched_get_active_thread(t, &by_waitpid);
+		Task* next = rec_sched_get_active_thread(*session,
+							 t, &by_waitpid);
 		if (!next) {
 			maybe_process_term_request(t);
 		}
 		t = next;
 
-		debug("line %d: Active task is %d. Events:",
-		      get_global_time(), t->tid);
+		LOG(debug) <<"line "<< t->trace_time() <<": Active task is "
+			   << t->tid <<". Events:";
 #ifdef DEBUGTAG
-		log_pending_events(t);
+		t->log_pending_events();
 #endif
 		int ptrace_event = t->ptrace_event();
-		assert_exec(t, (!by_waitpid || t->may_be_blocked() ||
-				ptrace_event),
-			    "%d unexpectedly runnable (0x%x) by waitpid",
-			    t->tid, t->status());
-		if (ptrace_event && !t->is_ptrace_seccomp_event()) {
-			handle_ptrace_event(&t);
-			if (!t) {
-				continue;
-			}
+		ASSERT(t, (!by_waitpid || t->may_be_blocked() ||
+			   ptrace_event))
+			<< "unexpectedly runnable ("<< HEX(t->status())
+			<<") by waitpid";
+		handle_ptrace_event(&t);
+		if (!t) {
+			continue;
+		}
+
+		if (t->unstable) {
+			// Do not record non-ptrace events for tasks in
+			// an unstable exit. We can't replay them.
+			LOG(debug) <<"Task in unstable exit; "
+				"refusing to record non-ptrace events";
+			continue;
 		}
 
 		bool did_initial_resume = false;
@@ -996,20 +1027,31 @@ void record(const char* rr_exe, int argc, char* argv[], char** envp)
 		runnable_state_changed(t);
 	}
 
-	log_info("Done recording -- cleaning up");
-	close_trace_files();
-	close_libpfm();
+	int exit_code = initial_task_group->exit_code;
+
+	LOG(info) <<"Done recording -- cleaning up";
+	session = nullptr;
+	initial_task_group = nullptr;
+
+	return exit_code;
 }
 
-void terminate_recording(Task* t)
+void terminate_recording(Task* t, int status)
 {
-	log_info("Processing termination request ...");
-	log_info("  recording final TRACE_TERMINATION event ...");
-	record_trace_termination_event(t);
-	flush_trace_files();
+	LOG(info) <<"Processing termination request ...";
+	LOG(info) <<"  recording final TRACE_TERMINATION event ...";
+
+	struct trace_frame frame;
+	memset(&frame, 0, sizeof(frame));
+	frame.tid = t ? t->tid : 0;
+	frame.global_time = session->ofstream().time();
+	frame.ev = Event(EV_TRACE_TERMINATION,
+			 BaseEvent(NO_EXEC_INFO)).encode();
+	session->ofstream() << frame;
+	session->ofstream().close();
 
 	// TODO: Task::killall() here?
 
-	log_info("  exiting, goodbye.");
-	exit(0);
+	LOG(info) <<"  exiting, goodbye.";
+	exit(status);
 }
