@@ -9,10 +9,13 @@ echo The slave is `uname -a`
 # ptrace hardening
 sudo ./src/script/setup.sh
 
+packages=(linux-libc-dev linux-libc-dev:i386
+	  gcc-multilib libc6-dev:i386 libdisasm-dev:i386 rpm
+	  g++ lib32stdc++6
+	  zlib1g:i386 zlib1g-dev:i386)
+
 sudo apt-get update && \
-    sudo apt-get install linux-libc-dev linux-libc-dev:i386 && \
-    sudo apt-get install gcc-multilib libc6-dev:i386 libdisasm-dev:i386 rpm && \
-    sudo apt-get install g++ lib32stdc++6 && \
+    sudo apt-get install "${packages[@]}"
     sudo ln -s /usr/lib32/libstdc++.so.6 /usr/lib32/libstdc++.so && \
     wget http://people.mozilla.org/~gal/libpfm_4.3.0-1_amd64.deb && \
     sudo dpkg -i libpfm_4.3.0-1_amd64.deb && \
