@@ -1,6 +1,26 @@
 /* -*- Mode: C; tab-width: 8; c-basic-offset: 8; indent-tabs-mode: t; -*- */
 
 #include "rrutil.h"
+#include<linux/net.h>
+
+#define CTRLMSG_LEN CMSG_LEN(sizeof(int))
+
+struct sendmmsg_arg
+{
+	int sockfd;
+	struct mmsghdr *msgvec;
+	unsigned int vlen;
+	unsigned int flags;
+};
+
+struct recvmmsg_arg
+{
+	int sockfd;
+	struct mmsghdr *msgvec;
+	unsigned int vlen;
+	unsigned int flags;
+	struct timespec *timeout;
+};
 
 static void breakpoint(void) {
 	int break_here = 1;
@@ -12,12 +32,6 @@ static int sockfds[2];
 
 static const int msg_magic = 0x1337beef;
 const ssize_t num_sockbuf_bytes = 1 << 20;
-/* TODO: rr doesn't know how to create scratch space for struct msghdr
- * yet, so it's forced to make the {send, recv}(m?)msg() functions
- * non-context-switchable.  This test would obviously deadlock in that
- * situation, so we help rr along by hiding the send->recv
- * synchronization behind a barrier, for now. */
-static pthread_barrier_t cheater_barrier;
 
 static void* reader_thread(void* dontcare) {
 	char token = '!';
@@ -36,7 +50,7 @@ static void* reader_thread(void* dontcare) {
 	for (i = 0; i < 2; ++i) {
 		atomic_puts("r: reading socket ...");
 		gettimeofday(&ts, NULL);
-		test_assert(1 == read(sock, &c, sizeof(c)));
+		check_syscall(1, read(sock, &c, sizeof(c)));
 		atomic_printf("r:   ... read '%c'\n", c);
 		test_assert(c == token);
 		++token;
@@ -45,13 +59,13 @@ static void* reader_thread(void* dontcare) {
 
 	atomic_puts("r: recv'ing socket ...");
 	gettimeofday(&ts, NULL);
-	test_assert(1 == recv(sock, &c, sizeof(c), 0));
+	check_syscall(1, recv(sock, &c, sizeof(c), 0));
 	atomic_printf("r:   ... recv'd '%c'\n", c);
 	test_assert(c == token);
 	++token;
 
 	atomic_puts("r: recvfrom'ing socket ...");
-	test_assert(1 == recvfrom(sock, &c, sizeof(c), 0, NULL, NULL));
+	check_syscall(1, recvfrom(sock, &c, sizeof(c), 0, NULL, NULL));
 	atomic_printf("r:   ... recvfrom'd '%c'\n", c);
 	test_assert(c == token);
 	++token;
@@ -60,8 +74,7 @@ static void* reader_thread(void* dontcare) {
 		socklen_t addrlen = sizeof(addr);
 
 		atomic_puts("r: recvfrom(&sock)'ing socket ...");
-		test_assert(1 == recvfrom(sock, &c, sizeof(c), 0,
-					  &addr, &addrlen));
+		check_syscall(1, recvfrom(sock, &c, sizeof(c), 0, &addr, &addrlen));
 		atomic_printf("r:   ... recvfrom'd '%c' from sock len:%d\n",
 			      c, addrlen);
 		test_assert(c == token);
@@ -81,12 +94,17 @@ static void* reader_thread(void* dontcare) {
 		mmsg.msg_hdr.msg_iov = &data;
 		mmsg.msg_hdr.msg_iovlen = 1;
 
+		struct cmsghdr *cmptr = (struct cmsghdr*)malloc(CTRLMSG_LEN);
+		mmsg.msg_hdr.msg_control = cmptr;
+		mmsg.msg_hdr.msg_controllen = CTRLMSG_LEN;
+
 		atomic_puts("r: recvmsg with DONTWAIT ...");
 		ret = recvmsg(sock, &mmsg.msg_hdr, MSG_DONTWAIT);
 		err = errno;
 		atomic_printf("r:  ... returned %d (%s/%d)\n",
 			      ret, strerror(err), err);
-		test_assert(-1 == ret && EWOULDBLOCK == err);
+		check_syscall(-1, ret);
+		test_assert(EWOULDBLOCK == err);
 
 		atomic_puts("r: recmsg'ing socket ...");
 
@@ -94,15 +112,40 @@ static void* reader_thread(void* dontcare) {
 		atomic_printf("r:   ... recvmsg'd 0x%x\n", magic);
 		test_assert(msg_magic == magic);
 
+		int fd = *(int *)CMSG_DATA(cmptr);
+		struct stat fs_new, fs_old;
+		fstat(fd, &fs_new);
+		fstat(STDERR_FILENO, &fs_old);
+		// check if control msg was send successfully
+		test_assert(fs_old.st_dev == fs_new.st_dev && fs_old.st_ino == fs_new.st_ino && fs_old.st_uid == fs_new.st_uid
+			 && fs_old.st_gid == fs_new.st_gid && fs_old.st_rdev == fs_new.st_rdev && fs_old.st_size == fs_new.st_size);
+
 		magic = ~msg_magic;
 		atomic_puts("r: recmmsg'ing socket ...");
 
-		pthread_barrier_wait(&cheater_barrier);
 		breakpoint();
-		test_assert(1 == recvmmsg(sock, &mmsg, 1, 0, NULL));
+		check_syscall(1, recvmmsg(sock, &mmsg, 1, 0, NULL));
 		atomic_printf("r:   ... recvmmsg'd 0x%x (%u bytes)\n",
 			      magic, mmsg.msg_len);
 		test_assert(msg_magic == magic);
+
+		magic = ~msg_magic;
+		struct recvmmsg_arg arg = {0};
+		arg.sockfd = sock;
+		arg.msgvec = &mmsg;
+		arg.vlen = 1;
+#if defined(SYS_socketcall)
+		check_syscall(1, syscall(SYS_socketcall, SYS_RECVMMSG, (void*)&arg));
+#elif defined(SYS_recvmmsg)
+		check_syscall(1, syscall(SYS_recvmmsg, (void*)&arg));
+#else
+#error unable to call recvmmsg
+#endif
+		atomic_printf("r:   ... recvmmsg'd(by socketcall) 0x%x (%u bytes)\n",
+			      magic, mmsg.msg_len);
+		test_assert(msg_magic == magic);
+
+		free(cmptr);
 	}
 	{
 		struct msghdr msg = { 0 };
@@ -118,7 +161,7 @@ static void* reader_thread(void* dontcare) {
 		msg.msg_iovlen = sizeof(iovs) / sizeof(iovs[0]);
 
 		atomic_puts("r: recmsg'ing socket with two iovs ...");
-		test_assert(2 == recvmsg(sock, &msg, 0));
+		check_syscall(2, recvmsg(sock, &msg, 0));
 		atomic_printf("r:   ... recvmsg'd '%c' and '%c'\n", c1, c2);
 
 		test_assert(c1 == token);
@@ -135,7 +178,7 @@ static void* reader_thread(void* dontcare) {
 		gettimeofday(&ts, NULL);
 		poll(&pfd, 1, -1);
 		atomic_puts("r:   ... done, doing nonblocking read ...");
-		test_assert(1 == read(sock, &c, sizeof(c)));
+		check_syscall(1, read(sock, &c, sizeof(c)));
 		atomic_printf("r:   ... read '%c'\n", c);
 		test_assert(c == token);
 		++token;
@@ -149,7 +192,7 @@ static void* reader_thread(void* dontcare) {
 		gettimeofday(&ts, NULL);
 		ppoll(&pfd, 1, NULL, NULL);
 		atomic_puts("r:   ... done, doing nonblocking read ...");
-		test_assert(1 == read(sock, &c, sizeof(c)));
+		check_syscall(1, read(sock, &c, sizeof(c)));
 		atomic_printf("r:   ... read '%c'\n", c);
 		test_assert(c == token);
 		++token;
@@ -166,12 +209,12 @@ static void* reader_thread(void* dontcare) {
 		ret = select(sock + 1, &fds, NULL, NULL, &tv);
 		atomic_printf("r:   ... returned %d; tv { %ld, %ld }\n",
 			      ret, tv.tv_sec, tv.tv_usec);
-		test_assert(1 == ret);
+		check_syscall(1, ret);
 		test_assert(FD_ISSET(sock, &fds));
 		test_assert(0 < tv.tv_sec && tv.tv_sec < infinity.tv_sec);
 
 		atomic_puts("r:   ... done, doing nonblocking read ...");
-		test_assert(1 == read(sock, &c, sizeof(c)));
+		check_syscall(1, read(sock, &c, sizeof(c)));
 		atomic_printf("r:   ... read '%c'\n", c);
 		test_assert(c == token);
 		++token;
@@ -185,12 +228,12 @@ static void* reader_thread(void* dontcare) {
 		ev.events = EPOLLIN;
 		ev.data.fd = sock;
 		gettimeofday(&ts, NULL);
-		test_assert(0 == epoll_ctl(epfd, EPOLL_CTL_ADD, ev.data.fd,
-					   &ev));
-		test_assert(1 == epoll_wait(epfd, &ev, 1, -1));
+		check_syscall(0, epoll_ctl(epfd, EPOLL_CTL_ADD, ev.data.fd, &ev));
+		check_syscall(1, epoll_wait(epfd, &ev, 1, -1));
 		atomic_puts("r:   ... done, doing nonblocking read ...");
 		test_assert(sock == ev.data.fd);
-		test_assert(1 == read(sock, &c, sizeof(c)));
+		check_syscall(1, epoll_wait(epfd, &ev, 1, -1));
+		check_syscall(1, read(sock, &c, sizeof(c)));
 		atomic_printf("r:   ... read '%c'\n", c);
 		test_assert(c == token);
 		++token;
@@ -205,13 +248,13 @@ static void* reader_thread(void* dontcare) {
 		++token;
 		memset(buf, token, num_sockbuf_bytes);
 
-		atomic_printf("r: writing outbuf of size %d ...\n",
+		atomic_printf("r: writing outbuf of size %zd ...\n",
 			      num_sockbuf_bytes);
 		while (nwritten < num_sockbuf_bytes) {
 			ssize_t this_write =
 				write(sock, buf,
 				      num_sockbuf_bytes - nwritten);
-			atomic_printf("r:   wrote %d bytes this time\n",
+			atomic_printf("r:   wrote %zd bytes this time\n",
 				      this_write);
 			nwritten += this_write;
 		}
@@ -220,11 +263,11 @@ static void* reader_thread(void* dontcare) {
 		memset(buf, token, num_sockbuf_bytes);
 		iov.iov_base = buf;
 		iov.iov_len = num_sockbuf_bytes;
-		atomic_printf("r: writev()ing outbuf of size %d ...\n",
+		atomic_printf("r: writev()ing outbuf of size %zd ...\n",
 			      num_sockbuf_bytes);
 		while (iov.iov_len > 0) {
 			ssize_t this_write = writev(sock, &iov, 1);
-			atomic_printf("r:   wrote %d bytes this time\n",
+			atomic_printf("r:   wrote %zd bytes this time\n",
 				      this_write);
 			iov.iov_len -= this_write;
 		}
@@ -236,11 +279,11 @@ static void* reader_thread(void* dontcare) {
 	{
 		sigset_t old_mask, mask;
 		sigfillset(&mask);
-		test_assert(0 == pthread_sigmask(SIG_BLOCK, &mask, &old_mask));
+		check_syscall(0, pthread_sigmask(SIG_BLOCK, &mask, &old_mask));
 
-		test_assert(1 == read(sock, &c, sizeof(c)));
+		check_syscall(1, read(sock, &c, sizeof(c)));
 
-		test_assert(0 == pthread_sigmask(SIG_SETMASK, &old_mask, NULL));
+		check_syscall(0, pthread_sigmask(SIG_SETMASK, &old_mask, NULL));
 	}
 	++token;
 	atomic_printf("r:   ... read '%c'\n", c);
@@ -262,7 +305,7 @@ static void read_all_chunks(int sock, char* buf, ssize_t num_sockbuf_bytes,
 					 num_sockbuf_bytes - nread);
 		int i;
 
-		atomic_printf("M:   read %d bytes this time,\n", this_read);
+		atomic_printf("M:   read %zd bytes this time,\n", this_read);
 		test_assert(this_read > 0);
 		/* XXX: we would like to assert that the written data
 		 * was read in more than one chunk, which should imply
@@ -279,7 +322,7 @@ static void read_all_chunks(int sock, char* buf, ssize_t num_sockbuf_bytes,
 			}
 		}
 		nread += this_read;
-		atomic_printf("M:      %d total so far\n", nread);
+		atomic_printf("M:      %zd total so far\n", nread);
 	}
 }
 
@@ -294,7 +337,6 @@ int main(int argc, char *argv[]) {
 	socketpair(AF_LOCAL, SOCK_STREAM, 0, sockfds);
 	sock = sockfds[0];
 
-	pthread_barrier_init(&cheater_barrier, NULL, 2);
 
 	pthread_mutex_lock(&lock);
 
@@ -311,7 +353,7 @@ int main(int argc, char *argv[]) {
 	atomic_puts("M: sleeping again ...");
 	usleep(500000);
 	atomic_printf("M: writing '%c' to socket ...\n", token);
-	test_assert(1 == write(sock, &token, sizeof(token)));
+	check_syscall(1, write(sock, &token, sizeof(token)));
 	++token;
 	atomic_puts("M:   ... done");
 	/* Force a wait on readv() */
@@ -322,7 +364,7 @@ int main(int argc, char *argv[]) {
 		atomic_puts("M: sleeping again ...");
 		usleep(500000);
 		atomic_printf("r: writev('%c')'ing socket ...\n", token);
-		test_assert(1 == writev(sock, &v, 1));
+		check_syscall(1, writev(sock, &v, 1));
 		++token;
 		atomic_puts("M:   ... done");
 	}
@@ -359,6 +401,14 @@ int main(int argc, char *argv[]) {
 		mmsg.msg_hdr.msg_iov = &data;
 		mmsg.msg_hdr.msg_iovlen = 1;
 
+		struct cmsghdr *cmptr = (struct cmsghdr*)malloc(CTRLMSG_LEN);// send a fd
+		cmptr->cmsg_level = SOL_SOCKET;
+		cmptr->cmsg_type = SCM_RIGHTS;
+		cmptr->cmsg_len = CTRLMSG_LEN;
+		mmsg.msg_hdr.msg_control = cmptr;
+		mmsg.msg_hdr.msg_controllen = CTRLMSG_LEN;
+		*(int *)CMSG_DATA(cmptr) = STDERR_FILENO; // send stderr as fd
+
 		/* Force a wait on recvmsg() */
 		atomic_puts("M: sleeping again ...");
 		usleep(500000);
@@ -377,7 +427,26 @@ int main(int argc, char *argv[]) {
 
 		sendmmsg(sock, &mmsg, 1, 0);
 		atomic_printf("M:   ... sent %u bytes\n", mmsg.msg_len);
-		pthread_barrier_wait(&cheater_barrier);
+
+		/* Force a wait on recvmmsg() */
+		atomic_puts("M: sleeping again ...");
+		usleep(500000);
+		atomic_printf("M: sendmmsg'ing(by socketcall) 0x%x to socket ...\n",
+			      msg_magic);
+
+		struct sendmmsg_arg arg = {0};
+		arg.sockfd = sock;
+		arg.msgvec = &mmsg;
+		arg.vlen = 1;
+#if defined(SYS_socketcall)
+		syscall(SYS_socketcall, SYS_SENDMMSG, (void*)&arg);
+#elif defined(SYS_sendmmsg)
+		syscall(SYS_sendmmsg, (void*)&arg);
+#else
+#error unable to call sendmmsg
+#endif
+
+		free(cmptr);
 	}
 	{
 		struct msghdr msg = { 0 };
@@ -398,7 +467,7 @@ int main(int argc, char *argv[]) {
 		usleep(500000);
 		atomic_printf("M: writing { '%c', '%c' } to socket ...\n",
 			      c1, c2);
-		test_assert(2 == sendmsg(sock, &msg, 0));
+		check_syscall(2, sendmsg(sock, &msg, 0));
 		atomic_puts("M:   ... done");
 	}
 	/* Force a wait on poll() */
