@@ -18,10 +18,15 @@ int main(int argc, char *argv[]) {
 	/* Make the value returned into |path| overlap two physical
 	 * pages. */
 	cwd = two_pages + pagesize - 3;
+	/* Fill pages with non-zeroes to ensure the returned string is
+	 * properly null-terminated */
+	memset(two_pages, 0xFF, two_pages_size);
 	test_assert(cwd == getcwd(cwd, two_pages_end - (void*)cwd));
 	atomic_printf("current working directory is %s; should be %s\n",
 		      cwd, expected_cwd);
 	test_assert(!strcmp(cwd, expected_cwd));
+	/* Make sure we didn't write too many bytes */
+	test_assert((unsigned char)cwd[strlen(cwd) + 1] == 0xFF);
 
 	atomic_puts("EXIT-SUCCESS");
 	return 0;
