@@ -9,7 +9,7 @@ static void breakpoint(void) {
 
 
 static void handle_sigtrap(int sig) {
-	atomic_puts("caught SIGTRAP!");
+	atomic_puts("EXIT-SUCCESS");
 	_exit(0);
 }
 
