@@ -40,7 +40,7 @@ int main(int argc, char *argv[]) {
 		atomic_printf("%d:%d,", rpage[i], wpage[i]);
 	}
 
-	atomic_puts(" done");
+	atomic_puts("EXIT-SUCCESS");
 
 	return 0;
 }
