@@ -2,8 +2,8 @@
 
 #include "rrutil.h"
 
-
 long int counter = 0;
+pthread_barrier_t bar;
 
 void catcher(int sig) {
 	atomic_printf("Signal caught, Counter is %ld\n", counter);
@@ -11,13 +11,14 @@ void catcher(int sig) {
 	_exit(0);
 }
 
-void* reciever(void* name) {
+void* receiver(void* name) {
 	struct sigaction sact;
 
 	sigemptyset(&sact.sa_mask);
 	sact.sa_flags = 0;
 	sact.sa_handler = catcher;
 	sigaction(SIGALRM, &sact, NULL);
+	pthread_barrier_wait(&bar);
 
 	while (1) {
 		counter++;
@@ -30,6 +31,7 @@ void* reciever(void* name) {
 
 void* sender(void* id) {
 	sleep(1);
+	pthread_barrier_wait(&bar);
 	pthread_kill(*((pthread_t*)id), SIGALRM);
 	return NULL;
 }
@@ -41,9 +43,11 @@ int main(void) {
 	/* (Kick on the syscallbuf lib.) */
 	gettimeofday(&tv, NULL);
 
+	/* init barrier */
+	pthread_barrier_init(&bar, NULL, 2);
 	/* Create independent threads each of which will execute
 	 * function */
-	pthread_create(&thread1, NULL, reciever, NULL);
+	pthread_create(&thread1, NULL, receiver, NULL);
 	pthread_create(&thread2, NULL, sender, &thread1);
 
 	/* Wait till threads are complete before main
